import unittest

from mindsdb.integrations.handlers.matrixone_handler.matrixone_handler import MatrixOneHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class MatrixOneHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "localhost",
                "port": 6001,
                "user": "dump",
                "password": "111",
                "database": "mo_catalog",
                "ssl": False
            }
        }
        cls.handler = MatrixOneHandler('test_mysql_handler', cls.kwargs)


    def test_0_connect(self):
        self.handler.connect()

    def test_1_drop_table(self):
        res = self.handler.query("DROP TABLE IF EXISTS PREM;")
        assert res.type is  not RESPONSE_TYPE.ERROR

    def test_2_create_table(self):
        res = self.handler.query("CREATE TABLE IF NOT EXISTS PREM (Premi varchar(50));")
        assert res.type is not  RESPONSE_TYPE.ERROR 

    def test_3_insert_table(self):
        res = self.handler.query("INSERT INTO PREM VALUES('Radha <3 Krishna');")
        assert res.type is not  RESPONSE_TYPE.ERROR 

    def test_4_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not  RESPONSE_TYPE.ERROR

    def test_5_select_query(self):
        query = "SELECT * FROM PREM;"
        result = self.handler.native_query(query)
        assert result.type is not  RESPONSE_TYPE.ERROR
    
    def test_6_get_columns(self):
        result = self.handler.get_columns('PREM')
        assert result.type is not  RESPONSE_TYPE.ERROR

    def test_7_check_connection(self):
        self.handler.check_connection()


if __name__ == '__main__':
    unittest.main()


import pprint

pydantic_schema_description = """## Understanding Pydantic Schemas for JSON Formatting

Pydantic schemas provide a framework for defining the structure and validation rules for JSON output. Below is an overview of key components commonly found in a Pydantic schema:

### Key Components

Each object in the schema represents a Pydantic model in JSON format. Typical fields in a Pydantic model description include:

- **`anyOf`**:
  - A list describing possible values for a Pydantic model field.

- **`additionalProperties`**:
  - Describes the keys of a dictionary. Keys are always of type `string` due to this being a JSON Pydantic schema. The corresponding key types supported by Pydantic are:
    - `string`: a text string
    - `integer`: an integer number
    - `number`: a floating-point number

- **`items`**:
  - Describes the items contained within an `array` (list).

- **`type`**:
  - Specifies the Pydantic type assigned to the field, defining the expected data type. Common types include:
    - `string`: a text string
    - `integer`: an integer number
    - `number`: a floating-point number
    - `array`: a list
    - `object`: a dictionary
    - `null`: the python null value None. Indicates the field is optional.

- **`description`**:
  - Provides a textual narrative explaining the purpose and details of the output JSON field.

- **`title`**:
  - A Pydantic-generated, human-readable title for the field.

- **`default`**:
  - The default value for this field if no value is provided by the user.

### Schema

Below is the Pydantic schema:

{schema}

### Examples

Below is an example of well-formed output adhering to this schema.

- Dummy text strings are represented as "lorem ipsum."

{example}
"""


def get_dummy_value(field_value):
    """A function to return a dummy value of a Pydantic model field."""
    type_str = field_value["type"]
    example_dict = {
        "string": "lorem ipsum",
        "int": 3,
        "number": 42.0,
        "null": None,
        "object": {"lorem ipsum": "lorem_ipsum"},
    }

    if type_str in example_dict:
        return example_dict[type_str]
    else:
        return None


def get_dummy_array(field_value):
    """A function to return a dummy array of a Pydantic model field."""
    items = field_value["items"]

    if "type" in items:
        if items["type"] == "null":  # skip if null
            pass
        elif items["type"] == "array":  # is it an array?
            array_value = get_dummy_array(items)
        elif (
            items["type"] == "object" and "additionalProperties" in items
        ):  # is it a dict?
            array_value = get_dummy_dict(items)
        else:  # it is a regular value!
            array_value = get_dummy_value(items)
        return [array_value for _ in range(2)]

    elif "AnyOf" in field_value["items"]:
        array_value = get_any_of(field_value["items"])  # can be one of many types
        return [array_value for _ in range(2)]

    else:  # is it a pydantic class?
        array_value = example_generator(items)
        return [array_value for _ in range(2)]


def get_dummy_dict(field_value):
    """A function to return a dummy dictionary of a Pydantic model field."""
    return get_dummy_value(field_value)


def get_any_of(field_value):
    """A function to return the first viable pydantic type of an Any() Pydantic model field."""
    for any_of in field_value["anyOf"]:
        if "type" in any_of:
            if any_of["type"] == "null":  # skip if null
                continue
            elif any_of["type"] == "array":  # is it an array?
                out = get_dummy_array(any_of)
                return out
            elif (
                any_of["type"] == "object" and "additionalProperties" in any_of
            ):  # is it a dict?
                out = get_dummy_dict(any_of)
                return out
            else:  # it is a regular value!
                out = get_dummy_value(any_of)
                return out
        else:  # is it a pydantic class?
            out = example_generator(any_of)
            return out


def example_generator(pydantic_json_schema):
    """dynamically parse a pydantic object and generate an example of it's formatting."""

    example_dict = {}
    for schema_name, schema in pydantic_json_schema.items():

        for field_name, field_value in schema.items():
            if "type" in field_value:

                if field_value["type"] == "array":  # is it an array?
                    example_dict[field_name] = get_dummy_array(field_value)

                elif (
                    field_value["type"] == "object"
                    and "additionalProperties" in field_value
                ):  # is it a dict?
                    example_dict[field_name] = get_dummy_dict(field_value)

                else:  # it is a regular value!
                    example_dict[field_name] = get_dummy_value(field_value)

            elif "anyOf" in field_value:
                example_dict[field_name] = get_any_of(field_value)

            else:  # it is a pydantic class
                example_dict[field_name] = example_generator(field_value)
    return example_dict


def search_and_replace_refs(schema, defs, ref_skip={}, n=0):
    """Dynamically substitute subclass references in a Pydantic object schema."""
    for key, value in schema.items():
        if key in ref_skip:
            continue
        if type(value) is dict:
            if "$ref" in value:
                definition_key = value["$ref"].split("/")[-1]
                if definition_key in ref_skip:
                    schema[key] = {"type": "null"}
                else:
                    schema[key] = {definition_key: defs[definition_key]["properties"]}
            else:
                search_and_replace_refs(value, defs, ref_skip, n + 1)
        elif type(value) is list:
            for val in value:
                search_and_replace_refs(val, defs, ref_skip, n + 1)


def remove_extraneous_fields(schema, ref_skip):
    """Remove extraneous fields from object descriptions."""
    reduced_schema = schema["properties"]

    for ref in ref_skip.keys():
        if ref in reduced_schema:
            del reduced_schema[ref]

    for key, value in reduced_schema.items():
        if "title" in value:
            del value["title"]
        if "$defs" in value:
            del value["$defs"]
        if "required" in value:
            del value["required"]

    return reduced_schema


def format_for_prompt(pydantic_object, ref_skip={}):
    """Format a Pydantic object description for prompting an LLM."""
    schema = {k: v for k, v in pydantic_object.schema().items()}

    search_and_replace_refs(
        schema=schema["properties"], defs=schema["$defs"], ref_skip=ref_skip, n=0
    )

    reduced_schema = remove_extraneous_fields(schema, ref_skip)

    reduced_schema = {schema["title"]: reduced_schema}

    out = pprint.pformat(reduced_schema)

    return out, reduced_schema



import pandas as pd
from mindsdb_sql_parser import ast
from pandas import DataFrame

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
)
from mindsdb.integrations.utilities.date_utils import parse_utc_date
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions


class SearchAnalyticsTable(APITable):
    """
    Table class for the Google Search Console Search Analytics table.
    """

    def select(self, query: ast.Select) -> DataFrame:
        """
        Gets all traffic data from the Search Console.

        Args:
            query (ast.Select): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.
        params = {}
        accepted_params = ['siteUrl', 'dimensions', 'type', 'rowLimit', 'aggregationType']
        for op, arg1, arg2 in conditions:
            if arg1 == 'startDate' or arg1 == 'endDate':
                date = parse_utc_date(arg2)
                if op == '=':
                    params[arg1] = date
                else:
                    raise NotImplementedError
            elif arg1 in accepted_params:
                if op != '=':
                    raise NotImplementedError
                params[arg1] = arg2
            else:
                raise NotImplementedError

        dimensions = ['query', 'page', 'device', 'country']

        # Get the group by from the query.
        params['dimensions'] = {}
        conditions = extract_comparison_conditions(query.group_by)
        for arg1 in conditions:
            if arg1 in dimensions:
                params['dimensions'][arg1] = arg1
            else:
                raise NotImplementedError

        # Get the order by from the query.
        if query.order_by is not None:
            if query.order_by[0].value == 'start_time':
                params['orderBy'] = 'startTime'
            elif query.order_by[0].value == 'updated':
                params['orderBy'] = 'updated'
            else:
                raise NotImplementedError

        if query.limit is not None:
            params['rowLimit'] = query.limit.value

        # Get the traffic data from the Google Search Console API.
        traffic_data = self.handler. \
            call_application_api(method_name='get_traffic_data', params=params)

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(traffic_data) == 0:
            traffic_data = pd.DataFrame([], columns=selected_columns)
        else:
            traffic_data.columns = self.get_columns()
            for col in set(traffic_data.columns).difference(set(selected_columns)):
                traffic_data = traffic_data.drop(col, axis=1)
        return traffic_data

    def get_columns(self) -> list:
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'keys',
            'clicks',
            'impressions',
            'ctr',
            'position'
        ]


class SiteMapsTable(APITable):
    """
    Table class for the Google Search Console Site Maps table.
    """

    def select(self, query: ast.Select) -> DataFrame:
        """
        Gets all traffic data from the Search Console.

        Args:
            query (ast.Select): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.
        params = {}
        accepted_params = ['siteUrl', 'sitemapIndex']
        for op, arg1, arg2 in conditions:
            if op != '=':
                raise NotImplementedError
            if arg1 in accepted_params:
                params[arg1] = arg2
            else:
                raise NotImplementedError

        if query.limit is not None:
            params['rowLimit'] = query.limit.value

        # Get the traffic data from the Google Search Console API.
        sitemaps = self.handler. \
            call_application_api(method_name='get_sitemaps', params=params)

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(sitemaps) == 0:
            sitemaps = pd.DataFrame([], columns=selected_columns)
        else:
            sitemaps.columns = self.get_columns()
            for col in set(sitemaps.columns).difference(set(selected_columns)):
                sitemaps = sitemaps.drop(col, axis=1)
        return sitemaps

    def insert(self, query: ast.Insert):
        """
        Submits a sitemap for a site.

        Args:
            query (ast.Insert): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Get the values from the query.
        values = query.values[0]
        params = {}
        # Get the event data from the values.
        for col, val in zip(query.columns, values):
            if col == 'siteUrl' or col == 'feedpath':
                params[col] = val
            else:
                raise NotImplementedError

        # Insert the event into the Google Calendar API.
        self.handler.call_application_api(method_name='submit_sitemap', params=params)

    def delete(self, query: ast.Delete):
        """
        Deletes a sitemap for a site.

        Args:
            query (ast.Delete): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.
        params = {}
        for op, arg1, arg2 in conditions:
            if op != '=':
                raise NotImplementedError
            if arg1 == 'siteUrl' or arg1 == 'feedpath':
                params[arg1] = arg2
            else:
                raise NotImplementedError

        # Delete the events in the Google Calendar API.
        self.handler.call_application_api(method_name='delete_sitemap', params=params)

    def get_columns(self) -> list:
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'path',
            'lastSubmitted',
            'isPending',
            'isSitemapsIndex',
            'type',
            'lastDownloaded',
            'warnings',
            'errors',
            'contents'
        ]



from http import HTTPStatus

from flask import request
from flask_restx import Resource

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics

from mindsdb.interfaces.jobs.jobs_controller import JobsController, parse_job_date


@ns_conf.route('/<project_name>/jobs')
class JobsResource(Resource):
    @ns_conf.doc('list_jobs')
    @api_endpoint_metrics('GET', '/jobs')
    def get(self, project_name):
        """ List all jobs in a project """
        jobs_controller = JobsController()

        return jobs_controller.get_list(project_name)

    @ns_conf.doc('create_job')
    @api_endpoint_metrics('POST', '/jobs')
    def post(self, project_name):
        """Create a job in a project"""

        # Check for required parameters.
        if 'job' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "job" parameter in POST body'
            )

        job = request.json['job']

        name = job.pop('name')
        if job['start_at'] is not None:
            job['start_at'] = parse_job_date(job['start_at'])
        if job['end_at'] is not None:
            job['end_at'] = parse_job_date(job['end_at'])

        jobs_controller = JobsController()

        create_job_name = jobs_controller.add(name, project_name, **job)

        return jobs_controller.get(create_job_name, project_name)


@ns_conf.route('/<project_name>/jobs/<job_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('job_name', 'Name of the job')
class JobResource(Resource):
    @ns_conf.doc('get_job')
    @api_endpoint_metrics('GET', '/jobs/job')
    def get(self, project_name, job_name):
        """Gets a job by name"""

        jobs_controller = JobsController()
        job_info = jobs_controller.get(job_name, project_name)
        if job_info is not None:
            return job_info

        return http_error(
            HTTPStatus.NOT_FOUND,
            'Job not found',
            f'Job with name {job_name} does not exist'
        )

    @ns_conf.doc('delete_job')
    @api_endpoint_metrics('DELETE', '/jobs/job')
    def delete(self, project_name, job_name):
        """Deletes a job by name"""

        jobs_controller = JobsController()
        jobs_controller.delete(job_name, project_name)

        return '', HTTPStatus.NO_CONTENT


@ns_conf.route('/<project_name>/jobs/<job_name>/history')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('job_name', 'Name of the job')
class JobsHistory(Resource):
    @ns_conf.doc('job_history')
    @api_endpoint_metrics('GET', '/jobs/job/history')
    def get(self, project_name, job_name):
        """Get history of job calls"""

        jobs_controller = JobsController()
        if jobs_controller.get(job_name, project_name) is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Job not found',
                f'Job with name {job_name} does not exist'
            )

        return jobs_controller.get_history(job_name, project_name)



import os
import typing

from mindsdb.utilities import log
from langfuse import Langfuse
from langfuse.client import StatefulSpanClient
from langfuse.callback import CallbackHandler
from langfuse.api.resources.commons.errors.not_found_error import NotFoundError as TraceNotFoundError

logger = log.getLogger(__name__)

# Define Langfuse public key.
LANGFUSE_PUBLIC_KEY = os.getenv("LANGFUSE_PUBLIC_KEY", "langfuse_public_key")

# Define Langfuse secret key.
LANGFUSE_SECRET_KEY = os.getenv("LANGFUSE_SECRET_KEY", "langfuse_secret_key")

# Define Langfuse host.
LANGFUSE_HOST = os.getenv("LANGFUSE_HOST", "http://localhost:3000")

# Define Langfuse environment.
LANGFUSE_ENVIRONMENT = os.getenv("LANGFUSE_ENVIRONMENT", "local")

# Define Langfuse release.
LANGFUSE_RELEASE = os.getenv("LANGFUSE_RELEASE", "local")

# Define Langfuse debug mode.
LANGFUSE_DEBUG = os.getenv("LANGFUSE_DEBUG", "false").lower() == "true"

# Define Langfuse timeout.
LANGFUSE_TIMEOUT = int(os.getenv("LANGFUSE_TIMEOUT", 10))

# Define Langfuse sample rate.
LANGFUSE_SAMPLE_RATE = float(os.getenv("LANGFUSE_SAMPLE_RATE", 1.0))

# Define if Langfuse is disabled.
LANGFUSE_DISABLED = os.getenv("LANGFUSE_DISABLED", "false").lower() == "true" or LANGFUSE_ENVIRONMENT == "local"
LANGFUSE_FORCE_RUN = os.getenv("LANGFUSE_FORCE_RUN", "false").lower() == "true"


class LangfuseClientWrapper:
    """
    Langfuse client wrapper. Defines Langfuse client configuration and initializes Langfuse client.
    """

    def __init__(self,
                 public_key: str = LANGFUSE_PUBLIC_KEY,
                 secret_key: str = LANGFUSE_SECRET_KEY,
                 host: str = LANGFUSE_HOST,
                 environment: str = LANGFUSE_ENVIRONMENT,
                 release: str = LANGFUSE_RELEASE,
                 debug: bool = LANGFUSE_DEBUG,
                 timeout: int = LANGFUSE_TIMEOUT,
                 sample_rate: float = LANGFUSE_SAMPLE_RATE,
                 disable: bool = LANGFUSE_DISABLED,
                 force_run: bool = LANGFUSE_FORCE_RUN) -> None:
        """
        Initialize Langfuse client.

        Args:
            public_key (str): Langfuse public key.
            secret_key (str): Langfuse secret key.
            host (str): Langfuse host.
            release (str): Langfuse release.
            timeout (int): Langfuse timeout.
            sample_rate (float): Langfuse sample rate.
        """

        self.metadata = None
        self.public_key = public_key
        self.secret_key = secret_key
        self.host = host
        self.environment = environment
        self.release = release
        self.debug = debug
        self.timeout = timeout
        self.sample_rate = sample_rate
        self.disable = disable
        self.force_run = force_run

        self.client = None
        self.trace = None
        self.metadata = None
        self.tags = None

        # Check if Langfuse is disabled.
        if LANGFUSE_DISABLED and not LANGFUSE_FORCE_RUN:
            logger.info("Langfuse is disabled.")
            return

        logger.info("Langfuse enabled")
        logger.debug(f"LANGFUSE_PUBLIC_KEY: {LANGFUSE_PUBLIC_KEY}")
        logger.debug(f"LANGFUSE_SECRET_KEY: {'*' * len(LANGFUSE_SECRET_KEY)}")
        logger.debug(f"LANGFUSE_HOST: {LANGFUSE_HOST}")
        logger.debug(f"LANGFUSE_ENVIRONMENT: {LANGFUSE_ENVIRONMENT}")
        logger.debug(f"LANGFUSE_RELEASE: {LANGFUSE_RELEASE}")
        logger.debug(f"LANGFUSE_DEBUG: {LANGFUSE_DEBUG}")
        logger.debug(f"LANGFUSE_TIMEOUT: {LANGFUSE_TIMEOUT}")
        logger.debug(f"LANGFUSE_SAMPLE_RATE: {LANGFUSE_SAMPLE_RATE * 100}%")

        self.client = Langfuse(
            public_key=public_key,
            secret_key=secret_key,
            host=host,
            release=release,
            debug=debug,
            timeout=timeout,
            sample_rate=sample_rate
        )

    def setup_trace(self,
                    name: str,
                    input: typing.Optional[typing.Any] = None,
                    tags: typing.Optional[typing.List] = None,
                    metadata: typing.Optional[typing.Dict] = None,
                    user_id: str = None,
                    session_id: str = None) -> None:
        """
        Setup trace. If Langfuse is disabled, nothing will be done.
        Args:
            name (str): Trace name.
            input (dict): Trace input.
            tags (dict): Trace tags.
            metadata (dict): Trace metadata.
            user_id (str): User ID.
            session_id (str): Session ID.
        """

        if self.client is None:
            logger.debug("Langfuse is disabled.")
            return

        self.set_metadata(metadata)
        self.set_tags(tags)

        try:
            self.trace = self.client.trace(
                name=name,
                input=input,
                metadata=self.metadata,
                tags=self.tags,
                user_id=user_id,
                session_id=session_id
            )
        except Exception as e:
            logger.error(f'Something went wrong while processing Langfuse trace {self.trace.id}: {str(e)}')

        logger.info(f"Langfuse trace configured with ID: {self.trace.id}")

    def get_trace_id(self) -> typing.Optional[str]:
        """
        Get trace ID. If Langfuse is disabled, returns None.
        """

        if self.client is None:
            logger.debug("Langfuse is disabled.")
            return ""

        if self.trace is None:
            logger.debug("Langfuse trace is not setup.")
            return ""

        return self.trace.id

    def start_span(self,
                   name: str,
                   input: typing.Optional[typing.Any] = None) -> typing.Optional[StatefulSpanClient]:
        """
        Create span. If Langfuse is disabled, nothing will be done.

        Args:
            name (str): Span name.
            input (dict): Span input.
        """

        if self.client is None:
            logger.debug("Langfuse is disabled.")
            return None

        return self.trace.span(name=name, input=input)

    def end_span_stream(self,
                        span: typing.Optional[StatefulSpanClient] = None) -> None:
        """
        End span. If Langfuse is disabled, nothing will happen.
        Args:
            span (Any): Span object.
        """

        if self.client is None:
            logger.debug("Langfuse is disabled.")
            return

        span.end()
        self.trace.update()

    def end_span(self,
                 span: typing.Optional[StatefulSpanClient] = None,
                 output: typing.Optional[typing.Any] = None) -> None:
        """
        End trace. If Langfuse is disabled, nothing will be done.

        Args:
            span (Any): Span object.
            output (Any): Span output.
        """

        if self.client is None:
            logger.debug("Langfuse is disabled.")
            return

        if span is None:
            logger.debug("Langfuse span is not created.")
            return

        span.end(output=output)
        self.trace.update(output=output)

        metadata = self.metadata or {}

        try:
            # Ensure all batched traces are sent before fetching.
            self.client.flush()
            metadata['tool_usage'] = self._get_tool_usage()
            self.trace.update(metadata=metadata)

        except Exception as e:
            logger.error(f'Something went wrong while processing Langfuse trace {self.trace.id}: {str(e)}')

    def get_langchain_handler(self) -> typing.Optional[CallbackHandler]:
        """
        Get Langchain handler. If Langfuse is disabled, returns None.
        """

        if self.client is None:
            logger.debug("Langfuse is disabled.")
            return None

        return self.trace.get_langchain_handler()

    def set_metadata(self, custom_metadata: dict = None) -> None:
        """
        Get default metadata.
        """
        self.metadata = custom_metadata or {}

        self.metadata["environment"] = self.environment
        self.metadata["release"] = self.release

    def set_tags(self, custom_tags: typing.Optional[typing.List] = None) -> None:
        """
        Get default tags.
        """
        self.tags = custom_tags or []

        self.tags.append(self.environment)
        self.tags.append(self.release)

    def _get_tool_usage(self) -> typing.Dict:
        """ Retrieves tool usage information from a langfuse trace.
        Note: assumes trace marks an action with string `AgentAction` """

        tool_usage = {}

        try:
            fetched_trace = self.client.get_trace(self.trace.id)
            steps = [s.name for s in fetched_trace.observations]
            for step in steps:
                if 'AgentAction' in step:
                    tool_name = step.split('-')[1]
                    if tool_name not in tool_usage:
                        tool_usage[tool_name] = 0
                    tool_usage[tool_name] += 1
        except TraceNotFoundError:
            logger.warning(f'Langfuse trace {self.trace.id} not found')
        except Exception as e:
            logger.error(f'Something went wrong while processing Langfuse trace {self.trace.id}: {str(e)}')

        return tool_usage



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    youtube_api_token={
        'type': ARG_TYPE.STR,
        'description': 'Youtube API Token',
        'label': 'Youtube API Token',
    },
    credentials_url={
        'type': ARG_TYPE.STR,
        'description': 'URL to Service Account Keys',
        'label': 'URL to Service Account Keys',
    },
    credentials_file={
        'type': ARG_TYPE.STR,
        'description': 'Location of Service Account Keys',
        'label': 'Path to Service Account Keys',
    },
    credentials={
        'type': ARG_TYPE.PATH,
        'description': 'Service Account Keys',
        'label': 'Upload Service Account Keys',
    },
    code={
        'type': ARG_TYPE.STR,
        'description': 'Code After Authorisation',
        'label': 'Code After Authorisation',
    },
)



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

import struct

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import SESSION_TRACK, SERVER_STATUS


class OkPacket(Packet):
    '''
    Implementation based on:
    https://mariadb.com/kb/en/library/1-connecting-connecting/#initial-handshake-packet
    '''

    '''
    int<1> 0x00 : OK_Packet header or (0xFE if CLIENT_DEPRECATE_EOF is set)
    int<lenenc> affected rows
    int<lenenc> last insert id
    int<2> server status
    int<2> warning count
    if session_tracking_supported (see CLIENT_SESSION_TRACK)
        string<lenenc> info
    if (status flags & SERVER_SESSION_STATE_CHANGED)
        string<lenenc> session state info
        string<lenenc> value of variable
    else
        string<EOF> info
    '''

    def setup(self):
        eof = self._kwargs.get('eof', False)
        self.ok_header = Datum('int<1>', 0xFE if eof is True else 0)
        self.affected_rows = Datum('int<lenenc>', self._kwargs.get('affected_rows') or 0)
        self.last_insert_id = Datum('int<lenenc>', 0)
        status = self._kwargs.get('status', 0x0002)
        self.server_status = Datum('int<2>', status)
        # Datum('int<2>', 0)
        self.warning_count = Datum('int<2>', 0)

        self.state_track = None
        state_track = self._kwargs.get('state_track')   # [[key: value]]
        if state_track is not None:
            accum = b''
            status = status | SERVER_STATUS.SERVER_SESSION_STATE_CHANGED
            self.server_status = Datum('int<2>', status)
            self.state_track = b''
            self.state_track += Datum('string<lenenc>', '').toStringPacket()  # 'info' - human readable status information
            for el in state_track:
                # NOTE at this moment just system variables
                name, value = el
                part = Datum('string<lenenc>', name).toStringPacket()
                part += Datum('string<lenenc>', value).toStringPacket()
                accum += struct.pack('i', SESSION_TRACK.SESSION_TRACK_SYSTEM_VARIABLES)[:1] + struct.pack('i', len(part))[:1] + part
                # self.state_track
                # self.state_track.append(Datum('string<lenenc>', '')
            accum = struct.pack('i', len(accum))[:1] + accum
            self.state_track += accum

        self.info = Datum('string<EOF>')

    @property
    def body(self):

        order = [
            'ok_header',
            'affected_rows',
            'last_insert_id',
            'server_status',
            'warning_count',
            'state_track',
            'info',
        ]
        string = b''
        for key in order:
            item = getattr(self, key)
            section_pack = b''
            if item is None:
                continue
            elif isinstance(item, bytes):
                section_pack = item
            else:
                section_pack = getattr(self, key).toStringPacket()
            string += section_pack

        self.setBody(string)
        return self._body

    @staticmethod
    def test():
        import pprint
        pprint.pprint(str(OkPacket(state_track=[['character_set_client', 'utf8'], ['character_set_connection', 'utf8'], ['character_set_results', 'utf8']]).get_packet_string()))


# only run the test if this file is called from debugger
if __name__ == "__main__":
    OkPacket.test()



import pymssql
from pymssql import OperationalError
import pandas as pd

from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import MYSQL_DATA_TYPE


logger = log.getLogger(__name__)


def _map_type(mssql_type_text: str) -> MYSQL_DATA_TYPE:
    """ Map MSSQL text types names to MySQL types as enum.

    Args:
        mssql_type_text (str): The name of the MSSQL type to map.

    Returns:
        MYSQL_DATA_TYPE: The MySQL type enum that corresponds to the MSSQL text type name.
    """
    internal_type_name = mssql_type_text.lower()
    types_map = {
        ('tinyint', 'smallint', 'int', 'bigint'): MYSQL_DATA_TYPE.INT,
        ('bit',): MYSQL_DATA_TYPE.BOOL,
        ('money', 'smallmoney', 'float', 'real'): MYSQL_DATA_TYPE.FLOAT,
        ('decimal', 'numeric'): MYSQL_DATA_TYPE.DECIMAL,
        ('date',): MYSQL_DATA_TYPE.DATE,
        ('time',): MYSQL_DATA_TYPE.TIME,
        ('datetime2', 'datetimeoffset', 'datetime', 'smalldatetime'): MYSQL_DATA_TYPE.DATETIME,
        ('varchar', 'nvarchar'): MYSQL_DATA_TYPE.VARCHAR,
        ('char', 'text', 'nchar', 'ntext'): MYSQL_DATA_TYPE.TEXT,
        ('binary', 'varbinary', 'image'): MYSQL_DATA_TYPE.BINARY
    }

    for db_types_list, mysql_data_type in types_map.items():
        if internal_type_name in db_types_list:
            return mysql_data_type

    logger.warning(f"MSSQL handler type mapping: unknown type: {internal_type_name}, use VARCHAR as fallback.")
    return MYSQL_DATA_TYPE.VARCHAR


class SqlServerHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Microsoft SQL Server statements.
    """
    name = 'mssql'

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.connection_args = kwargs.get('connection_data')
        self.dialect = 'mssql'
        self.database = self.connection_args.get('database')
        self.renderer = SqlalchemyRender('mssql')

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Establishes a connection to a Microsoft SQL Server database.

        Raises:
            pymssql._mssql.OperationalError: If an error occurs while connecting to the Microsoft SQL Server database.

        Returns:
            pymssql.Connection: A connection object to the Microsoft SQL Server database.
        """

        if self.is_connected is True:
            return self.connection

        # Mandatory connection parameters
        if not all(key in self.connection_args for key in ['host', 'user', 'password', 'database']):
            raise ValueError('Required parameters (host, user, password, database) must be provided.')

        config = {
            'host': self.connection_args.get('host'),
            'user': self.connection_args.get('user'),
            'password': self.connection_args.get('password'),
            'database': self.connection_args.get('database')
        }

        # Optional connection parameters
        if 'port' in self.connection_args:
            config['port'] = self.connection_args.get('port')

        if 'server' in self.connection_args:
            config['server'] = self.connection_args.get('server')

        try:
            self.connection = pymssql.connect(**config)
            self.is_connected = True
            return self.connection
        except OperationalError as e:
            logger.error(f'Error connecting to Microsoft SQL Server {self.database}, {e}!')
            self.is_connected = False
            raise

    def disconnect(self):
        """
        Closes the connection to the Microsoft SQL Server database if it's currently open.
        """

        if not self.is_connected:
            return
        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the Microsoft SQL Server database.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            with connection.cursor() as cur:
                # Execute a simple query to test the connection
                cur.execute('select 1;')
            response.success = True
        except OperationalError as e:
            logger.error(f'Error connecting to Microsoft SQL Server {self.database}, {e}!')
            response.error_message = str(e)

        if response.success and need_to_close:
            self.disconnect()
        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Executes a SQL query on the Microsoft SQL Server database and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor(as_dict=True) as cur:
            try:
                cur.execute(query)
                if cur.description:
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK, affected_rows=cur.rowcount)
                connection.commit()
            except Exception as e:
                logger.error(f'Error running query: {query} on {self.database}, {e}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_code=0,
                    error_message=str(e)
                )
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """

        query_str = self.renderer.get_string(query, with_failback=True)
        logger.debug(f"Executing SQL query: {query_str}")
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables and views in the current schema of the Microsoft SQL Server database.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """

        query = f"""
            SELECT
                table_schema,
                table_name,
                table_type
            FROM {self.database}.INFORMATION_SCHEMA.TABLES
            WHERE TABLE_TYPE in ('BASE TABLE', 'VIEW');
        """
        return self.native_query(query)

    def get_columns(self, table_name) -> Response:
        """
        Retrieves column details for a specified table in the Microsoft SQL Server database.

        Args:
            table_name (str): The name of the table for which to retrieve column information.

        Returns:
            Response: A response object containing the column details, formatted as per the `Response` class.
        Raises:
            ValueError: If the 'table_name' is not a valid string.
        """

        query = f"""
            SELECT
                column_name as "Field",
                data_type as "Type"
            FROM
                information_schema.columns
            WHERE
                table_name = '{table_name}'
        """
        result = self.native_query(query)
        if result.resp_type is RESPONSE_TYPE.TABLE:
            result.data_frame['mysql_data_type'] = result.data_frame['Type'].apply(_map_type)
        return result



from typing import Any, Dict, Union, Optional, List
from uuid import uuid4
import datetime
import json

from langchain_core.callbacks.base import BaseCallbackHandler

from mindsdb.utilities import log
from mindsdb.interfaces.storage import db

logger = log.getLogger(__name__)
logger.setLevel('DEBUG')


class LangfuseCallbackHandler(BaseCallbackHandler):
    """Langchain callback handler that traces tool & chain executions using Langfuse."""

    def __init__(self, langfuse, trace_id: Optional[str] = None, observation_id: Optional[str] = None):
        self.langfuse = langfuse
        self.chain_uuid_to_span = {}
        self.action_uuid_to_span = {}
        # if these are not available, we generate some UUIDs
        self.trace_id = trace_id or uuid4().hex
        self.observation_id = observation_id or uuid4().hex
        # Track metrics about tools and chains
        self.tool_metrics = {}
        self.chain_metrics = {}
        self.current_chain = None

    def on_tool_start(
            self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        """Run when tool starts running."""
        parent_run_uuid = kwargs.get('parent_run_id', uuid4()).hex
        action_span = self.action_uuid_to_span.get(parent_run_uuid)
        if action_span is None:
            return

        tool_name = serialized.get("name", "tool")
        start_time = datetime.datetime.now()

        # Initialize or update tool metrics
        if tool_name not in self.tool_metrics:
            self.tool_metrics[tool_name] = {
                'count': 0,
                'total_time': 0,
                'errors': 0,
                'last_error': None,
                'inputs': []
            }

        self.tool_metrics[tool_name]['count'] += 1
        self.tool_metrics[tool_name]['inputs'].append(input_str)

        metadata = {
            'tool_name': tool_name,
            'started': start_time.isoformat(),
            'start_timestamp': start_time.timestamp(),
            'input_length': len(input_str) if input_str else 0
        }
        action_span.update(metadata=metadata)

    def on_tool_end(self, output: str, **kwargs: Any) -> Any:
        """Run when tool ends running."""
        parent_run_uuid = kwargs.get('parent_run_id', uuid4()).hex
        action_span = self.action_uuid_to_span.get(parent_run_uuid)
        if action_span is None:
            return

        end_time = datetime.datetime.now()
        tool_name = action_span.metadata.get('tool_name', 'unknown')
        start_timestamp = action_span.metadata.get('start_timestamp')

        if start_timestamp:
            duration = end_time.timestamp() - start_timestamp
            if tool_name in self.tool_metrics:
                self.tool_metrics[tool_name]['total_time'] += duration

        metadata = {
            'finished': end_time.isoformat(),
            'duration_seconds': duration if start_timestamp else None,
            'output_length': len(output) if output else 0
        }

        action_span.update(
            output=output,  # tool output is action output (unless superseded by a global action output)
            metadata=metadata
        )

    def on_tool_error(
            self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        """Run when tool errors."""
        parent_run_uuid = kwargs.get('parent_run_id', uuid4()).hex
        action_span = self.action_uuid_to_span.get(parent_run_uuid)
        if action_span is None:
            return

        try:
            error_str = str(error)
        except Exception:
            error_str = "Couldn't get error string."

        tool_name = action_span.metadata.get('tool_name', 'unknown')
        if tool_name in self.tool_metrics:
            self.tool_metrics[tool_name]['errors'] += 1
            self.tool_metrics[tool_name]['last_error'] = error_str

        metadata = {
            'error_description': error_str,
            'error_type': error.__class__.__name__,
            'error_time': datetime.datetime.now().isoformat()
        }
        action_span.update(metadata=metadata)

    def on_chain_start(
            self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> Any:
        """Run when chain starts running."""
        if self.langfuse is None:
            return

        run_uuid = kwargs.get('run_id', uuid4()).hex

        if serialized is None:
            serialized = {}

        chain_name = serialized.get("name", "chain")
        start_time = datetime.datetime.now()

        # Initialize or update chain metrics
        if chain_name not in self.chain_metrics:
            self.chain_metrics[chain_name] = {
                'count': 0,
                'total_time': 0,
                'errors': 0,
                'last_error': None
            }

        self.chain_metrics[chain_name]['count'] += 1
        self.current_chain = chain_name

        try:
            chain_span = self.langfuse.span(
                name=f'{chain_name}-{run_uuid}',
                trace_id=self.trace_id,
                parent_observation_id=self.observation_id,
                input=json.dumps(inputs, indent=2)
            )

            metadata = {
                'chain_name': chain_name,
                'started': start_time.isoformat(),
                'start_timestamp': start_time.timestamp(),
                'input_keys': list(inputs.keys()) if isinstance(inputs, dict) else None,
                'input_size': len(inputs) if isinstance(inputs, dict) else len(str(inputs))
            }
            chain_span.update(metadata=metadata)
            self.chain_uuid_to_span[run_uuid] = chain_span
        except Exception as e:
            logger.warning(f"Error creating Langfuse span: {str(e)}")

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> Any:
        """Run when chain ends running."""
        if self.langfuse is None:
            return

        chain_uuid = kwargs.get('run_id', uuid4()).hex
        if chain_uuid not in self.chain_uuid_to_span:
            return
        chain_span = self.chain_uuid_to_span.pop(chain_uuid)
        if chain_span is None:
            return

        try:
            end_time = datetime.datetime.now()
            chain_name = chain_span.metadata.get('chain_name', 'unknown')
            start_timestamp = chain_span.metadata.get('start_timestamp')

            if start_timestamp and chain_name in self.chain_metrics:
                duration = end_time.timestamp() - start_timestamp
                self.chain_metrics[chain_name]['total_time'] += duration

            metadata = {
                'finished': end_time.isoformat(),
                'duration_seconds': duration if start_timestamp else None,
                'output_keys': list(outputs.keys()) if isinstance(outputs, dict) else None,
                'output_size': len(outputs) if isinstance(outputs, dict) else len(str(outputs))
            }
            chain_span.update(output=json.dumps(outputs, indent=2), metadata=metadata)
            chain_span.end()
        except Exception as e:
            logger.warning(f"Error updating Langfuse span: {str(e)}")

    def on_chain_error(self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any) -> Any:
        """Run when chain errors."""
        chain_uuid = kwargs.get('run_id', uuid4()).hex
        if chain_uuid not in self.chain_uuid_to_span:
            return
        chain_span = self.chain_uuid_to_span.get(chain_uuid)
        if chain_span is None:
            return

        try:
            error_str = str(error)
        except Exception:
            error_str = "Couldn't get error string."

        chain_name = chain_span.metadata.get('chain_name', 'unknown')
        if chain_name in self.chain_metrics:
            self.chain_metrics[chain_name]['errors'] += 1
            self.chain_metrics[chain_name]['last_error'] = error_str

        metadata = {
            'error_description': error_str,
            'error_type': error.__class__.__name__,
            'error_time': datetime.datetime.now().isoformat()
        }
        chain_span.update(metadata=metadata)

    def on_agent_action(self, action, **kwargs: Any) -> Any:
        """Run on agent action."""
        if self.langfuse is None:
            return

        run_uuid = kwargs.get('run_id', uuid4()).hex
        try:
            action_span = self.langfuse.span(
                name=f'{getattr(action, "type", "action")}-{getattr(action, "tool", "")}-{run_uuid}',
                trace_id=self.trace_id,
                parent_observation_id=self.observation_id,
                input=str(action)
            )
            self.action_uuid_to_span[run_uuid] = action_span
        except Exception as e:
            logger.warning(f"Error creating Langfuse span for agent action: {str(e)}")

    def on_agent_finish(self, finish, **kwargs: Any) -> Any:
        """Run on agent end."""
        if self.langfuse is None:
            return

        run_uuid = kwargs.get('run_id', uuid4()).hex
        if run_uuid not in self.action_uuid_to_span:
            return
        action_span = self.action_uuid_to_span.pop(run_uuid)
        if action_span is None:
            return

        try:
            if finish is not None:
                action_span.update(output=finish)  # supersedes tool output
            action_span.end()
        except Exception as e:
            logger.warning(f"Error updating Langfuse span: {str(e)}")

    def auth_check(self):
        if self.langfuse is not None:
            return self.langfuse.auth_check()
        return False

    def get_metrics(self) -> Dict[str, Any]:
        """Get collected metrics about tools and chains.

        Returns:
            Dict containing:
            - tool_metrics: Statistics about tool usage, errors, and timing
            - chain_metrics: Statistics about chain execution, errors, and timing
            For each tool/chain, includes:
                - count: Number of times used
                - total_time: Total execution time
                - errors: Number of errors
                - last_error: Most recent error message
                - avg_duration: Average execution time
        """
        metrics = {
            'tool_metrics': {},
            'chain_metrics': {}
        }

        # Process tool metrics
        for tool_name, data in self.tool_metrics.items():
            metrics['tool_metrics'][tool_name] = {
                'count': data['count'],
                'total_time': data['total_time'],
                'avg_duration': data['total_time'] / data['count'] if data['count'] > 0 else 0,
                'errors': data['errors'],
                'last_error': data['last_error'],
                'error_rate': data['errors'] / data['count'] if data['count'] > 0 else 0
            }

        # Process chain metrics
        for chain_name, data in self.chain_metrics.items():
            metrics['chain_metrics'][chain_name] = {
                'count': data['count'],
                'total_time': data['total_time'],
                'avg_duration': data['total_time'] / data['count'] if data['count'] > 0 else 0,
                'errors': data['errors'],
                'last_error': data['last_error'],
                'error_rate': data['errors'] / data['count'] if data['count'] > 0 else 0
            }

        return metrics


def get_skills(agent: db.Agents) -> List:
    """ Retrieve skills from agent `skills` attribute. Specific to agent endpoints. """
    return [rel.skill.type for rel in agent.skills_relationships]



import requests
from urllib.parse import urljoin

from mindsdb.utilities import log

logger = log.getLogger(__name__)


def move_under(d, key_contents_to_move, key_to_move_under=None):
    """
    Moves the keys in nested dict with key key_contents_to_move under the dict defined by key_to_move_under.
    If no key_to_move_under is provided, the keys are moved under d.
    eg.
    Calling this on the following dict (d) with key_contents_to_move = "a":
    {
        "a": {
            "b": 1,
            "c": 2
        }
    }
    results in:
    {
        "b": 1,
        "c": 2
    }
    """
    if key_contents_to_move not in d:
        return
    for k, v in d[key_contents_to_move].items():
        if key_to_move_under:
            d[key_to_move_under][k] = v
        else:
            d[k] = v
    del d[key_contents_to_move]


class SAPERP:

    def __init__(self, url: str, api_key: str) -> None:
        self.base_url = url
        self.api_key = api_key

    def _request(self, method: str, relative_endpoint: str, data=None):
        kwargs = {
            "method": method,
            "url": urljoin(self.base_url, relative_endpoint),
            "headers": {
                "APIKey": self.api_key,
                "Accept": "application/json",
                "DataServiceVersion": "2.0"
            }
        }
        if data is not None:
            kwargs["data"] = data
        return requests.request(**kwargs)

    def is_connected(self) -> bool:
        if self._request("get", "").ok:
            return True
        return False

    def get(self, endpoint):
        """ Common method for all get endpoints """
        try:
            resp = self._request("get", endpoint)
            if resp.ok:
                resp = resp.json()["d"]
                if "results" in resp:
                    resp = resp["results"]
                else:
                    resp = [resp]
            else:
                resp = []
            for r in resp:
                move_under(r, "__metadata")
            return resp
        except Exception as e:
            logger.error(f"Error requesting endpoint {endpoint}: {e}")
            return {}



from abc import ABC, abstractmethod
from typing import Text, List, Dict

from requests.exceptions import RequestException

from mindsdb.integrations.utilities.handlers.api_utilities.microsoft.ms_graph_api_utilities import MSGraphAPIBaseClient
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class MSGraphAPITeamsClient(MSGraphAPIBaseClient, ABC):
    """
    The base class for the Microsoft Graph API client for the Microsoft Teams handler.
    """

    def check_connection(self) -> bool:
        """
        Check if the connection to Microsoft Teams is established.

        Returns:
            bool: True if the connection is established, False otherwise.
        """
        try:
            self._get_all_groups()
            return True
        except RequestException as request_error:
            logger.error(f"Failed to check connection to Microsoft Teams: {request_error}")
            return False
        
    def get_teams(self) -> List[Dict]:
        """
        Get teams from Microsoft Teams.

        Returns:
            List[Dict]: The teams data.
        """
        return self._get_all_groups()
        
    def get_channels(self, group_id: Text = None, channel_ids: List[Text] = None) -> List[Dict]:
        """
        Get channels from Microsoft Teams.

        Args:
            group_id (Text): The ID of the group that the channels belong to.
            channel_ids (List[Text]): The IDs of the channels.

        Returns:
            List[Dict]: The channels data.
        """
        if group_id and channel_ids:
            return self._get_channels_in_group_by_ids(group_id, channel_ids)
        elif group_id:
            return self._get_all_channels_in_group(group_id)
        elif channel_ids:
            return self._get_channels_across_all_groups_by_ids(channel_ids)
        else:
            return self._get_all_channels_across_all_groups()
        
    def get_channel_messages(self, group_id: Text, channel_id: Text, message_ids: List[Text] = None) -> List[Dict]:
        """
        Get channel messages from Microsoft Teams.

        Args:
            group_id (Text): The ID of the group that the channel belongs to.
            channel_id (Text): The ID of the channel that the messages belong to.
            message_ids (List[Text]): The IDs of the messages.

        Returns:
            List[Dict]: The messages data.
        """
        if message_ids:
            return self._get_messages_in_channel_by_ids(group_id, channel_id, message_ids)
        else:
            return self._get_all_messages_in_channel(group_id, channel_id)
        
    def get_chats(self, chat_ids: List[Text] = None) -> List[Dict]:
        """
        Get chats from Microsoft Teams.

        Args:
            chat_ids (List[Text]): The IDs of the chats.

        Returns:
            List[Dict]: The chats data.
        """
        if chat_ids:
            return self._get_chats_by_ids(chat_ids)
        else:
            return self._get_all_chats()
        
    def get_chat_messages(self, chat_id: Text, message_ids: List[Text] = None) -> List[Dict]:
        """
        Get chat messages from Microsoft Teams.

        Args:
            chat_id (Text): The ID of the chat that the messages belong to.
            message_ids (List[Text]): The IDs of the messages.

        Returns:
            List[Dict]: The messages data.
        """
        if message_ids:
            return self._get_messages_in_chat_by_ids(chat_id, message_ids)
        else:
            return self._get_all_messages_in_chat(chat_id)

    def _get_all_group_ids(self) -> List[Text]:
        """
        Get all group IDs related to Microsoft Teams.

        Returns:
            List[Text]: The group IDs.
        """
        if not self._group_ids:
            groups = self._get_all_groups()
            self._group_ids = [group["id"] for group in groups]

        return self._group_ids
    
    @abstractmethod
    def _get_all_groups(self) -> List[Dict]:
        """
        Get all groups related to Microsoft Teams.

        Returns:
            List[Dict]: The groups data.
        """
        pass

    @abstractmethod
    def _get_chat_by_id(self, chat_id: Text) -> Dict:
        """
        Get a chat by its ID.

        Args:
            chat_id (Text): The ID of the chat.

        Returns:
            Dict: The chat data.
        """
        pass

    @abstractmethod
    def _get_all_chats(self, limit: int = None) -> List[Dict]:
        """
        Get all chats related to Microsoft Teams.

        Args:
            limit (int): The maximum number of chats to return.

        Returns:
            List[Dict]: The chats data.
        """
        pass

    @abstractmethod
    def _get_message_in_chat_by_id(self, chat_id: Text, message_id: Text) -> Dict:
        """
        Get a message by its ID and the ID of the chat that it belongs to.

        Args:
            chat_id (Text): The ID of the chat that the message belongs to.
            message_id (Text): The ID of the message.

        Returns:
            Dict: The message data.
        """
        pass

    @abstractmethod
    def _get_all_messages_in_chat(self, chat_id: Text, limit: int = None) -> List[Dict]:
        """
        Get messages of a chat by its ID.

        Args:
            chat_id (Text): The ID of the chat.

        Returns:
            List[Dict]: The messages data.
        """
        pass

    def _get_channel_in_group_by_id(self, group_id: Text, channel_id: Text) -> Dict:
        """
        Get a channel by its ID and the ID of the group that it belongs to.

        Args:
            group_id (Text): The ID of the group that the channel belongs to.
            channel_id (Text): The ID of the channel.

        Returns:
            Dict: The channel data.
        """
        channel = self.fetch_data_json(f"teams/{group_id}/channels/{channel_id}")
        # Add the group ID to the channel data.
        channel.update({"teamId": group_id})

        return channel
    
    def _get_channels_in_group_by_ids(self, group_id: Text, channel_ids: List[Text]) -> List[Dict]:
        """
        Get channels by their IDs and the ID of the group that they belong to.

        Args:
            group_id (Text): The ID of the group that the channels belong to.
            channel_ids (List[Text]): The IDs of the channels.

        Returns:
            List[Dict]: The channels data.
        """
        channels = []
        for channel_id in channel_ids:
            channels.append(self._get_channel_in_group_by_id(group_id, channel_id))

        return channels
        
    def _get_all_channels_in_group(self, group_id: Text) -> List[Dict]:
        """
        Get all channels of a group by its ID.

        Args:
            group_id (Text): The ID of the group.

        Returns:
            List[Dict]: The channels data.
        """
        channels = self.fetch_data_json(f"teams/{group_id}/channels")
        for channel in channels:
            channel["teamId"] = group_id

        return channels
    
    def _get_all_channels_across_all_groups(self) -> List[Dict]:
        """
        Get all channels across all groups related to Microsoft Teams.

        Returns:
            List[Dict]: The channels data.
        """
        channels = []
        for group_id in self._get_all_group_ids():
            channels += self._get_all_channels_in_group(group_id)

        return channels
    
    def _get_channels_across_all_groups_by_ids(self, channel_ids: List[Text]) -> List[Dict]:
        """
        Get channels by their IDs.

        Args:
            channel_ids (List[Text]): The IDs of the channels.

        Returns:
            List[Dict]: The channels data.
        """
        channels = self._get_all_channels_across_all_groups()

        return [channel for channel in channels if channel["id"] in channel_ids]
    
    def _get_message_in_channel_by_id(self, group_id: Text, channel_id: Text, message_id: Text) -> Dict:
        """
        Get a message by its ID, the ID of the group that it belongs to, and the ID of the channel that it belongs to.

        Args:
            group_id (Text): The ID of the group that the channel belongs to.
            channel_id (Text): The ID of the channel that the message belongs to.
            message_id (Text): The ID of the message.

        Returns:
            Dict: The message data.
        """
        return self.fetch_data_json(f"teams/{group_id}/channels/{channel_id}/messages/{message_id}")
    
    def _get_messages_in_channel_by_ids(self, group_id: Text, channel_id: Text, message_ids: List[Text]) -> List[Dict]:
        """
        Get messages by their IDs, the ID of the group that they belong to, and the ID of the channel that they belong to.

        Args:
            group_id (Text): The ID of the group that the channel belongs to.
            channel_id (Text): The ID of the channel that the messages belong to.
            message_ids (List[Text]): The IDs of the messages.

        Returns:
            List[Dict]: The messages data.
        """
        messages = []
        for message_id in message_ids:
            messages.append(self._get_message_in_channel_by_id(group_id, channel_id, message_id))

        return messages

    def _get_all_messages_in_channel(self, group_id: Text, channel_id: Text, limit: int = None) -> List[Dict]:
        """
        Get messages of a channel by its ID and the ID of the group that it belongs to.

        Args:
            group_id (Text): The ID of the group that the channel belongs to.
            channel_id (Text): The ID of the channel.

        Returns:
            List[Dict]: The messages data.
        """
        messages = []
        for messages_batch in self.fetch_paginated_data(f"teams/{group_id}/channels/{channel_id}/messages"):
            messages += messages_batch

            if limit and len(messages) >= limit:
                break

        return messages[:limit]
    
    def _get_chats_by_ids(self, chat_ids: List[Text]) -> List[Dict]:
        """
        Get chats by their IDs.

        Args:
            chat_ids (List[Text]): The IDs of the chats.

        Returns:
            List[Dict]: The chats data.
        """
        chats = []
        for chat_id in chat_ids:
            chats.append(self._get_chat_by_id(chat_id))

        return chats
    
    def _get_messages_in_chat_by_ids(self, chat_id: Text, message_ids: List[Text]) -> List[Dict]:
        """
        Get messages by their IDs and the ID of the chat that they belong to.

        Args:
            chat_id (Text): The ID of the chat that the messages belong to.
            message_ids (List[Text]): The IDs of the messages.

        Returns:
            List[Dict]: The messages data.
        """
        messages = []
        for message_id in message_ids:
            messages.append(self._get_message_in_chat_by_id(chat_id, message_id))

        return messages


class MSGraphAPITeamsApplicationPermissionsClient(MSGraphAPITeamsClient):
    """
    The Microsoft Graph API client for the Microsoft Teams handler with application permissions.
    This client is used for accessing the Microsoft Teams specific endpoints of the Microsoft Graph API.
    Several common methods for submitting requests, fetching data, etc. are inherited from the base class.
    """
        
    def _get_all_groups(self) -> List[Dict]:
        """
        Get all groups related to Microsoft Teams.

        Returns:
            List[Dict]: The groups data.
        """
        return self.fetch_data_json(
            "/groups",
            params={"$filter": "resourceProvisioningOptions/Any(x:x eq 'Team')"}
        )
    
    def _get_chat_by_id(self, chat_id: Text) -> Dict:
        """
        Get a chat by its ID.

        Args:
            chat_id (Text): The ID of the chat.

        Returns:
            Dict: The chat data.
        """
        return self.fetch_data_json(f"chats/{chat_id}")
    
    def _get_all_chats(self, limit: int = None) -> List[Dict]:
        """
        Get all chats related to Microsoft Teams.

        Args:
            limit (int): The maximum number of chats to return.

        Returns:
            List[Dict]: The chats data.
        """
        raise RuntimeError("Retrieving all chats is not supported with application permissions. Either use delegated permissions or provide a chat ID.")

    def _get_message_in_chat_by_id(self, chat_id: Text, message_id: Text) -> Dict:
        """
        Get a message by its ID and the ID of the chat that it belongs to.

        Args:
            chat_id (Text): The ID of the chat that the message belongs to.
            message_id (Text): The ID of the message.

        Returns:
            Dict: The message data.
        """
        return self.fetch_data_json(f"chats/{chat_id}/messages/{message_id}")
    
    def _get_all_messages_in_chat(self, chat_id: Text, limit: int = None) -> List[Dict]:
        """
        Get messages of a chat by its ID.

        Args:
            chat_id (Text): The ID of the chat.

        Returns:
            List[Dict]: The messages data.
        """
        messages = []
        for messages_batch in self.fetch_paginated_data(f"chats/{chat_id}/messages"):
            messages += messages_batch

            if limit and len(messages) >= limit:
                break

        return messages[:limit]


class MSGraphAPITeamsDelegatedPermissionsClient(MSGraphAPITeamsClient): 
    """
    The Microsoft Graph API client for the Microsoft Teams handler with delegated permissions.
    This client is used for accessing the Microsoft Teams specific endpoints of the Microsoft Graph API.
    Several common methods for submitting requests, fetching data, etc. are inherited from the base class.
    """
        
    def _get_all_groups(self) -> List[Dict]:
        """
        Get all groups that the signed in user is a member of.

        Returns:
            List[Dict]: The groups data.
        """
        return self.fetch_data_json("me/joinedTeams")
    
    def _get_chat_by_id(self, chat_id: Text) -> Dict:
        """
        Get a chat by its ID.

        Args:
            chat_id (Text): The ID of the chat.

        Returns:
            Dict: The chat data.
        """
        return self.fetch_data_json(f"/me/chats/{chat_id}")

    def _get_all_chats(self, limit: int = None) -> List[Dict]:
        """
        Get all chats of the signed in user.

        Args:
            limit (int): The maximum number of chats to return.

        Returns:
            List[Dict]: The chats data.
        """
        chats = []
        for chat_batch in self.fetch_paginated_data("me/chats"):
            chats += chat_batch

            if limit and len(chats) >= limit:
                break

        return chats[:limit]
    
    def _get_message_in_chat_by_id(self, chat_id: Text, message_id: Text) -> Dict:
        """
        Get a message by its ID and the ID of the chat that it belongs to.

        Args:
            chat_id (Text): The ID of the chat that the message belongs to.
            message_id (Text): The ID of the message.

        Returns:
            Dict: The message data.
        """
        return self.fetch_data_json(f"me/chats/{chat_id}/messages/{message_id}")
    
    def _get_all_messages_in_chat(self, chat_id: Text, limit: int = None) -> List[Dict]:
        """
        Get messages of a chat by its ID.

        Args:
            chat_id (Text): The ID of the chat.

        Returns:
            List[Dict]: The messages data.
        """
        messages = []
        for messages_batch in self.fetch_paginated_data(f"me/chats/{chat_id}/messages"):
            messages += messages_batch

            if limit and len(messages) >= limit:
                break

        return messages[:limit]



from pyod.models.ecod import ECOD  # unsupervised default
from pyod.models.xgbod import XGBOD  # semi-supervised default
from catboost import CatBoostClassifier  # supervised default


def train_unsupervised(X_train, model=None):
    model = model if model is not None else ECOD()
    model.fit(X_train)
    return model


def train_semisupervised(X_train, y_train):
    model = XGBOD(estimator_list=[ECOD()])
    model.fit(X_train, y_train)
    return model


def train_supervised(X_train, y_train, model=None):
    model = model if model is not None else CatBoostClassifier(logging_level="Silent")
    model.fit(X_train, y_train)
    return model



from mindsdb.integrations.handlers.strava_handler.strava_tables import StravaAllClubsTable
from mindsdb.integrations.handlers.strava_handler.strava_tables import StravaClubActivitesTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql

import requests
import pandas as pd
import json
from collections import OrderedDict
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from stravalib.client import Client

logger = log.getLogger(__name__)

class StravaHandler(APIHandler):
    """Strava handler implementation"""

    def __init__(self, name=None, **kwargs):
        """Initialize the Strava handler.
        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})

        self.parser = parse_sql
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.connection = None
        self.is_connected = False

        strava_all_clubs_data = StravaAllClubsTable(self)
        self._register_table("all_clubs", strava_all_clubs_data)
        
        strava_club_activites_data = StravaClubActivitesTable(self)
        self._register_table("club_activities", strava_club_activites_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        if self.is_connected is True:
            return self.connection

        client = Client()
        url = client.authorization_url(client_id=self.connection_data['strava_client_id'], redirect_uri='http://127.0.0.1:5000/authorization')
        client.access_token = self.connection_data['strava_access_token']
        self.connection = client

        return self.connection

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.
        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to Strava API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query)
        return self.query(ast)


connection_args = OrderedDict(
    strava_client_id={
        'type': ARG_TYPE.STR,
        'description': 'Client id for accessing Strava Application API'
    },
    strava_access_token={
        'type': ARG_TYPE.STR,
        'description': 'Access Token for accessing Strava Application API'
    }
)

connection_args_example = OrderedDict(
    strava_client_id ='<your-strava-client_id>',
    strava_access_token ='<your-strava-api-token>'
)



from pandas import DataFrame

import sqlanydb
import sqlalchemy_sqlany.base as sqlany_dialect

from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


logger = log.getLogger(__name__)

class SQLAnyHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the SAP SQL Anywhere statements.
    """

    name = 'sqlany'

    def __init__(self, name: str, connection_data: dict, **kwargs):
        super().__init__(name)

        self.dialect = 'sqlany'
        self.parser = parse_sql
        self.connection_data = connection_data
        self.renderer = SqlalchemyRender(sqlany_dialect.SQLAnyDialect)
        self.host = self.connection_data.get('host')
        self.port = self.connection_data.get('port')
        self.userid = self.connection_data.get('user')
        self.password = self.connection_data.get('password')
        self.server = self.connection_data.get('server')
        self.databaseName = self.connection_data.get('database')
        self.encryption = self.connection_data.get('encrypt', False)
        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Handles the connection to a SAP SQL Anywhere database insance.
        """

        if self.is_connected is True:
            return self.connection

        if self.port.strip().isnumeric():
            self.host += ":" + self.port.strip()

        if self.encryption:
            self.encryption = "SIMPLE"
        else:
            self.encryption = "NONE"
        
        connection = sqlanydb.connect(
            host=self.host,
            userid=self.userid,
            password=self.password,
            server=self.server,
            databaseName=self.databaseName,
            encryption=self.encryption
        )
        self.is_connected = True
        self.connection = connection
        return self.connection

    def disconnect(self):
        """
        Disconnects from the SAP SQL Anywhere database
        """

        if self.is_connected is True:
            self.connection.close()
            self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the SAP SQL Anywhere database
        :return: success status and error message if error occurs
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            cur = connection.cursor()
            cur.execute('SELECT 1 FROM SYS.DUMMY;')
            response.success = True
        except sqlanydb.Error as e:
            logger.error(f'Error connecting to SAP SQL Anywhere {self.host}, {e}!')
            response.error_message = e

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in SAP SQL Anywhere
        :return: returns the records from the current recordset
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        cur = connection.cursor()
        try:
            cur.execute(query)
            if not cur.description:
                response = Response(RESPONSE_TYPE.OK)
            else:
                result = cur.fetchall()
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    DataFrame(
                        result,
                        columns=[x[0] for x in cur.description]
                    )
                )
            connection.commit()
        except Exception as e:
            logger.error(f'Error running query: {query} on {self.connection}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_code=0,
                error_message=str(e)
            )
            connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """

        query_str = self.renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        List all tables in SAP SQL Anywhere in the current schema
        """

        return self.native_query(f"""
            SELECT USER_NAME(ob.UID) AS SCHEMA_NAME
            , st.TABLE_NAME
            , st.TABLE_TYPE
            FROM SYSOBJECTS ob
            INNER JOIN SYS.SYSTABLE st on ob.ID = st.OBJECT_ID
            WHERE ob.TYPE='U' AND st.TABLE_TYPE <> 'GBL TEMP'
        """)

    def get_columns(self, table_name: str) -> Response:
        """
        List all columns in a table in SAP SQL Anywhere in the current schema
        :param table_name: the table name for which to list the columns
        :return: returns the columns in the table
        """

        return self.renderer.dialect.get_columns(table_name)



import pandas as pd
import threading
import queue
from typing import List

from mindsdb_sql_parser import ASTNode
from mindsdb.api.executor.planner.steps import FetchDataframeStepPartition
from mindsdb.integrations.utilities.query_traversal import query_traversal

from mindsdb.interfaces.query_context.context_controller import RunningQuery
from mindsdb.api.executor.sql_query.result_set import ResultSet
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import Context, context as ctx
from mindsdb.utilities.partitioning import get_max_thread_count, split_data_frame
from mindsdb.api.executor.sql_query.steps.fetch_dataframe import get_table_alias, get_fill_param_fnc

from .base import BaseStepCall


logger = log.getLogger(__name__)


class FetchDataframePartitionCall(BaseStepCall):
    """
    Alternative to FetchDataframeCall but fetch data by batches wrapping user's query to:

     select * from ({user query})
      where {track_column} > {previous value}
      order by track_column
      limit size {batch_size} `

    """

    bind = FetchDataframeStepPartition

    def call(self, step: FetchDataframeStepPartition) -> ResultSet:
        """
        Parameters:
        - batch_size - count of rows to fetch from database per iteration, optional default 1000
        - threads - run partitioning in threads, bool or int, optinal, if set:
           - int value: use this as count of threads
           - true: table threads, autodetect count of thread
           - false: disable threads even if ml task queue is enabled
        - track_column - column used for creating partitions
          - query will be sorted by this column and select will be limited by batch_size
        - error (default raise)
          - when `error='skip'`, errors in partition will be skipped and execution will be continued
        """

        self.dn = self.session.datahub.get(step.integration)
        query = step.query

        # fill params
        fill_params = get_fill_param_fnc(self.steps_data)
        query_traversal(query, fill_params)

        # get query record
        run_query = self.sql_query.run_query
        if run_query is None:
            raise RuntimeError('Error with partitioning of the query')
        run_query.set_params(step.params)

        self.table_alias = get_table_alias(step.query.from_table, self.context.get('database'))
        self.current_step_num = step.step_num
        self.substeps = step.steps

        config = Config()

        # ml task queue enabled?
        use_threads, thread_count = False, None
        if config['ml_task_queue']['type'] == 'redis':
            use_threads = True

        # use threads?
        if 'threads' in step.params:
            threads = step.params['threads']
            if isinstance(threads, int):
                thread_count = threads
                use_threads = True
            if threads is True:
                use_threads = True
            if threads is False:
                # disable even with ml task queue
                use_threads = False

        on_error = step.params.get('error', 'raise')
        if use_threads:
            return self.fetch_threads(run_query, query, thread_count=thread_count, on_error=on_error)
        else:
            return self.fetch_iterate(run_query, query, on_error=on_error)

    def fetch_iterate(self, run_query: RunningQuery, query: ASTNode, on_error: str = None) -> ResultSet:
        """
         Process batches one by one in circle
        """

        results = []
        while True:

            # fetch batch
            query2 = run_query.get_partition_query(self.current_step_num, query)
            response = self.dn.query(
                query=query2,
                session=self.session
            )
            df = response.data_frame

            if df is None or len(df) == 0:
                break

            # executing of sub steps can modify dataframe columns, lets memorise max tracking value
            max_track_value = run_query.get_max_track_value(df)
            try:
                sub_data = self.exec_sub_steps(df)
                results.append(sub_data)
            except Exception as e:
                if on_error == 'skip':
                    logger.error(e)
                else:
                    raise e

            run_query.set_progress(df, max_track_value)

        return self.concat_results(results)

    def concat_results(self, results: List[ResultSet]) -> ResultSet:
        """
        Concatenate list of result sets to single result set
        """
        df_list = []
        for res in results:
            df, col_names = res.to_df_cols()
            if len(df) > 0:
                df_list.append(df)

        data = ResultSet()
        if len(df_list) > 0:
            data.from_df_cols(pd.concat(df_list), col_names)

        return data

    def exec_sub_steps(self, df: pd.DataFrame) -> ResultSet:
        """
        FetchDataframeStepPartition has substeps defined
        Every batch of data have to be used to execute these substeps
        - batch of data is put as result of FetchDataframeStepPartition
        - substep are executed using result of previos step (like it is all fetched data is available)
        - the final result is returned and used outside to concatenate with results of other's batches
        """

        input_data = ResultSet()

        input_data.from_df(
            df,
            table_name=self.table_alias[1],
            table_alias=self.table_alias[2],
            database=self.table_alias[0]
        )

        # execute with modified previous results
        steps_data2 = self.steps_data.copy()
        steps_data2[self.current_step_num] = input_data

        sub_data = None
        for substep in self.substeps:
            sub_data = self.sql_query.execute_step(substep, steps_data=steps_data2)
            steps_data2[substep.step_num] = sub_data
        return sub_data

    def fetch_threads(self, run_query: RunningQuery, query: ASTNode,
                      thread_count: int = None, on_error: str = None) -> ResultSet:
        """
        Process batches in threads
        - spawn required count of threads
        - create in/out queue to communicate with threads
        - send task to threads and receive results
        """

        # create communication queues
        queue_in = queue.Queue()
        queue_out = queue.Queue()
        self.stop_event = threading.Event()

        if thread_count is None:
            thread_count = get_max_thread_count()

        # 3 tasks per worker during 1 batch
        partition_size = int(run_query.batch_size / thread_count / 3)
        # min partition size
        if partition_size < 10:
            partition_size = 10

        # create N workers pool
        workers = []
        results = []

        try:
            for i in range(thread_count):
                worker = threading.Thread(target=self._worker, daemon=True, args=(ctx.dump(), queue_in,
                                                                                  queue_out, self.stop_event))
                worker.start()
                workers.append(worker)

            while True:
                # fetch batch
                query2 = run_query.get_partition_query(self.current_step_num, query)
                response = self.dn.query(
                    query=query2,
                    session=self.session
                )
                df = response.data_frame

                if df is None or len(df) == 0:
                    # TODO detect circles: data handler ignores condition and output is repeated

                    # exit & stop workers
                    break

                max_track_value = run_query.get_max_track_value(df)

                # split into chunks and send to workers
                sent_chunks = 0
                for df2 in split_data_frame(df, partition_size):
                    queue_in.put([sent_chunks, df2])
                    sent_chunks += 1

                batch_results = []
                for i in range(sent_chunks):
                    res = queue_out.get()
                    if 'error' in res:
                        if on_error == 'skip':
                            logger.error(res['error'])
                        else:
                            raise RuntimeError(res['error'])

                    if res['data']:
                        batch_results.append(res)

                # sort results
                batch_results.sort(key=lambda x: x['num'])

                results.append(self.concat_results(
                    [item['data'] for item in batch_results]
                ))

                # TODO
                #  1. get next batch without updating track_value:
                #    it allows to keep queue_in filled with data between fetching batches
                run_query.set_progress(df, max_track_value)
        finally:
            self.close_workers(workers)

        return self.concat_results(results)

    def close_workers(self, workers: List[threading.Thread]):
        """
        Sent signal to workers to stop
        """

        self.stop_event.set()
        for worker in workers:
            if worker.is_alive():
                worker.join()

    def _worker(self, context: Context, queue_in: queue.Queue, queue_out: queue.Queue, stop_event: threading.Event):
        """
        Worker function. Execute incoming tasks unless stop_event is set
        """
        ctx.load(context)
        while True:
            if stop_event.is_set():
                break

            try:
                chunk_num, df = queue_in.get(timeout=1)
                if df is None:
                    continue

                sub_data = self.exec_sub_steps(df)

                queue_out.put({'data': sub_data, 'num': chunk_num})
            except queue.Empty:
                continue

            except Exception as e:
                queue_out.put({'error': str(e)})
                stop_event.set()



import json
from typing import Dict, List
from pathlib import Path

from mindsdb.utilities import log
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.interfaces.storage.fs import FileStorageFactory, RESOURCE_GROUP, FileStorage


logger = log.getLogger(__name__)


TABS_FILENAME = 'tabs'


def get_storage():
    # deprecated

    storageFactory = FileStorageFactory(
        resource_group=RESOURCE_GROUP.TAB,
        sync=True
    )

    # resource_id is useless for 'tabs'
    # use constant
    return storageFactory(0)


class TabsController:
    """Tool for adding, editing, and deleting user's tabs

    Attributes:
        storage_factory (FileStorageFactory): callable object which returns tabs file storage
    """

    def __init__(self) -> None:
        self.storage_factory = FileStorageFactory(
            resource_group=RESOURCE_GROUP.TAB,
            sync=True
        )

    def _get_file_storage(self) -> FileStorage:
        """Get user's tabs file storage
           NOTE: file storage depend is company_id sensitive, so need to recreate it each time

        Returns:
            FileStorage
        """
        return self.storage_factory(0)

    def _get_next_tab_id(self) -> int:
        """Get next free tab id

        Returns:
            int: id for next tab
        """
        tabs_files = self._get_tabs_files()
        tabs_ids = list(tabs_files.keys())
        if len(tabs_ids) == 0:
            return 1
        return max(tabs_ids) + 1

    def _get_tabs_files(self) -> Dict[int, Path]:
        """Get list of paths to each tab file

        Returns:
            Dict[int, Path]
        """
        tabs = {}
        for child in self._get_file_storage().folder_path.iterdir():
            if (child.is_file() and child.name.startswith('tab_')) is False:
                continue
            tab_id = child.name.replace('tab_', '')
            if tab_id.isnumeric() is False:
                continue
            tabs[int(tab_id)] = child
        return tabs

    def _get_tabs_meta(self) -> List[Dict]:
        """Get tabs info without content

        Returns:
            List[Dict]
        """
        all_tabs = self.get_all()
        for tab in all_tabs:
            del tab['content']
        return all_tabs

    def _migrate_legacy(self) -> None:
        """Convert old single-file tabs storage to multiple files
        """
        file_storage = self._get_file_storage()
        try:
            file_data = file_storage.file_get(TABS_FILENAME)
        except FileNotFoundError:
            return
        except Exception:
            file_storage.delete()
            return

        try:
            data = json.loads(file_data)
        except Exception:
            file_storage.delete()
            return

        if isinstance(data, dict) is False or isinstance(data.get('tabs'), str) is False:
            file_storage.delete()
            return

        try:
            tabs_list = json.loads(data['tabs'])
        except Exception:
            file_storage.delete()
            return

        if isinstance(tabs_list, list) is False:
            file_storage.delete()
            return

        for tab in tabs_list:
            tab_id = self._get_next_tab_id()

            b_types = json.dumps({
                'index': tab.get('index', 0),
                'name': tab.get('name', 'undefined'),
                'content': tab.get('value', '')
            }).encode("utf-8")
            file_storage.file_set(f'tab_{tab_id}', b_types)

        file_storage.delete(TABS_FILENAME)

    def get_all(self) -> List[Dict]:
        """Get list of all tabs

        Returns:
            List[Dict]: all tabs data
        """
        self._get_file_storage().pull()
        self._migrate_legacy()

        tabs_files = self._get_tabs_files()
        tabs_list = []
        for tab_id, tab_path in tabs_files.items():
            try:
                data = json.loads(tab_path.read_text())
            except Exception as e:
                logger.error(f"Can't read data of tab {ctx.company_id}/{tab_id}: {e}")
                continue
            tabs_list.append({
                'id': tab_id,
                **data
            })

        tabs_list.sort(key=lambda x: x['index'])
        return tabs_list

    def get(self, tab_id: int) -> Dict:
        """Get data of single tab

        Args:
            tab_id (int): id of the tab

        Returns:
            dict: tabs data
        """
        if isinstance(tab_id, int) is False:
            raise ValueError('Tab id must be integer')

        try:
            raw_tab_data = self._get_file_storage().file_get(f'tab_{tab_id}')
        except FileNotFoundError:
            raise EntityNotExistsError(f'tab {tab_id}')

        try:
            data = json.loads(raw_tab_data)
        except Exception as e:
            logger.error(f"Can't read data of tab {ctx.company_id}/{tab_id}: {e}")
            raise Exception(f"Can't read data of tab: {e}")

        return {
            'id': tab_id,
            **data
        }

    def add(self, index: int = None, name: str = 'undefined', content: str = '') -> Dict:
        """Add new tab

        Args:
            index (int, optional): index of new tab
            name (str, optional): name of new tab
            content (str, optional): content of new tab

        Returns:
            dict: new tab meta info: id, name and index
        """
        file_storage = self._get_file_storage()
        tab_id = self._get_next_tab_id()

        reorder_required = index is not None
        if index is None:
            all_tabs = self.get_all()
            if len(all_tabs) == 0:
                index = 0
            else:
                index = max([x.get('index', 0) for x in all_tabs]) + 1

        data_bytes = json.dumps({
            'index': index,
            'name': name,
            'content': content
        }).encode("utf-8")
        file_storage.file_set(f'tab_{tab_id}', data_bytes)

        if reorder_required:
            all_tabs = self.get_all()
            all_tabs.sort(key=lambda x: (x['index'], 0 if x['id'] == tab_id else 1))
            file_storage.sync = False
            for tab_index, tab in enumerate(all_tabs):
                tab['index'] = tab_index
                data_bytes = json.dumps(tab).encode('utf-8')
                file_storage.file_set(f'tab_{tab["id"]}', data_bytes)
            file_storage.sync = True
            file_storage.push()

        return {'id': tab_id, 'index': index, 'name': name}

    def modify(self, tab_id: int, index: int = None, name: str = None, content: str = None) -> Dict:
        """Modify the tab

        Args:
            tab_id (int): if of the tab to modify
            index (int, optional): tab's new index
            name (str, optional): tab's new name
            content (str, optional): tab's new content

        Returns:
            dict: new tab meta info: id, name and index
        """
        file_storage = self._get_file_storage()
        current_data = self.get(tab_id)

        # region modify index
        if index is not None and current_data['index'] != index:
            current_data['index'] = index
            all_tabs = [x for x in self.get_all() if x['id'] != tab_id]
            all_tabs.sort(key=lambda x: x['index'])
            file_storage.sync = False
            for tab_index, tab in enumerate(all_tabs):
                if tab_index < index:
                    tab['index'] = tab_index
                else:
                    tab['index'] = tab_index + 1
                data_bytes = json.dumps(tab).encode('utf-8')
                file_storage.file_set(f'tab_{tab["id"]}', data_bytes)
            file_storage.sync = True
            file_storage.push()
        # endregion

        # region modify name
        if name is not None and current_data['name'] != name:
            current_data['name'] = name
        # endregion

        # region modify content
        if content is not None and current_data['content'] != content:
            current_data['content'] = content
        # endregion

        data_bytes = json.dumps(current_data).encode('utf-8')
        file_storage.file_set(f'tab_{tab_id}', data_bytes)

        return {
            'id': current_data['id'],
            'index': current_data['index'],
            'name': current_data['name']
        }

    def delete(self, tab_id: int):
        file_storage = self._get_file_storage()
        try:
            file_storage.file_get(f'tab_{tab_id}')
        except FileNotFoundError:
            raise EntityNotExistsError(f'tab {tab_id}')

        file_storage.delete(f'tab_{tab_id}')


tabs_controller = TabsController()




class QueryPlan:
    def __init__(self, steps=None, **kwargs):
        self.steps = []
        self.is_resumable = False

        if steps:
            for step in steps:
                self.add_step(step)

    def __eq__(self, other):
        if type(self) != type(other):
            return False

        if len(self.steps) != len(other.steps):
            return False

        for step, other_step in zip(self.steps, other.steps):
            if step != other_step:
                return False

    @property
    def last_step_index(self):
        return len(self.steps) - 1

    def add_step(self, step):
        if not step.step_num:
            step.step_num = len(self.steps)
        self.steps.append(step)
        return self.steps[-1]



import os
import tempfile
import importlib
import multipart
from pathlib import Path
from http import HTTPStatus

from flask import request, send_file, current_app as ca
from flask_restx import Resource

from mindsdb_sql_parser.ast import Identifier
from mindsdb_sql_parser.ast.mindsdb import CreateMLEngine

from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.integrations.libs.ml_exec_base import process_cache
from mindsdb.integrations.utilities.install import install_dependencies
from mindsdb.interfaces.storage.model_fs import HandlerStorage
from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.handlers import ns_conf
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.executor.command_executor import ExecuteCommands
from mindsdb.utilities.config import Config


@ns_conf.route('/')
class HandlersList(Resource):
    @ns_conf.doc('handlers_list')
    @api_endpoint_metrics('GET', '/handlers')
    def get(self):
        '''List all db handlers'''

        if request.args.get('lazy') == '1':
            handlers = ca.integration_controller.get_handlers_metadata()
        else:
            handlers = ca.integration_controller.get_handlers_import_status()
        result = []
        for handler_type, handler_meta in handlers.items():
            # remove non-integration handlers
            if handler_type not in ['utilities', 'dummy_data']:
                row = {'name': handler_type}
                row.update(handler_meta)
                del row['path']
                result.append(row)
        return result


@ns_conf.route('/<handler_name>/icon')
class HandlerIcon(Resource):
    @ns_conf.param('handler_name', 'Handler name')
    @api_endpoint_metrics('GET', '/handlers/handler/icon')
    def get(self, handler_name):
        try:
            handler_meta = ca.integration_controller.get_handlers_metadata().get(handler_name)
            if handler_meta is None:
                return http_error(HTTPStatus.NOT_FOUND, 'Icon not found', f'Icon for {handler_name} not found')
            icon_name = handler_meta['icon']['name']
            handler_folder = handler_meta['import']['folder']
            mindsdb_path = Path(importlib.util.find_spec('mindsdb').origin).parent
            icon_path = mindsdb_path.joinpath('integrations/handlers').joinpath(handler_folder).joinpath(icon_name)
            if icon_path.is_absolute() is False:
                icon_path = Path(os.getcwd()).joinpath(icon_path)
        except Exception:
            return http_error(HTTPStatus.NOT_FOUND, 'Icon not found', f'Icon for {handler_name} not found')
        else:
            return send_file(icon_path)


@ns_conf.route('/<handler_name>')
class HandlerInfo(Resource):
    @ns_conf.param('handler_name', 'Handler name')
    @api_endpoint_metrics('GET', '/handlers/handler')
    def get(self, handler_name):

        handler_meta = ca.integration_controller.get_handler_meta(handler_name)
        row = {'name': handler_name}
        row.update(handler_meta)
        del row['path']
        del row['icon']
        return row


@ns_conf.route('/<handler_name>/install')
class InstallDependencies(Resource):
    @ns_conf.param('handler_name', 'Handler name')
    @api_endpoint_metrics('POST', '/handlers/handler/install')
    def post(self, handler_name):
        handler_meta = ca.integration_controller.get_handler_meta(handler_name)

        if handler_meta is None:
            return f'Unknown handler: {handler_name}', 400

        if handler_meta.get('import', {}).get('success', False) is True:
            return 'Installed', 200

        dependencies = handler_meta['import']['dependencies']
        if len(dependencies) == 0:
            return 'Installed', 200

        result = install_dependencies(dependencies)

        # reload it if any result, so we can get new error message
        ca.integration_controller.reload_handler_module(handler_name)
        if result.get('success') is True:
            # If warm processes are available in the cache, remove them.
            # This will force a new process to be created with the installed dependencies.
            process_cache.remove_processes_for_handler(handler_name)
            return '', 200
        return http_error(
            500,
            f'Failed to install dependencies for {handler_meta.get("title", handler_name)}',
            result.get('error_message', 'unknown error')
        )


def prepare_formdata():
    params = {}
    file_names = []

    def on_field(field):
        name = field.field_name.decode()
        value = field.value.decode()
        params[name] = value

    def on_file(file):
        params[file.field_name.decode()] = file.file_object
        file_names.append(file.field_name.decode())

    temp_dir_path = temp_dir_path = tempfile.mkdtemp(
        prefix='mindsdb_byom_file_',
        dir=Config().paths['tmp']
    )

    parser = multipart.create_form_parser(
        headers=request.headers,
        on_field=on_field,
        on_file=on_file,
        config={
            'UPLOAD_DIR': temp_dir_path.encode(),  # bytes required
            'UPLOAD_KEEP_FILENAME': True,
            'UPLOAD_KEEP_EXTENSIONS': True,
            'MAX_MEMORY_FILE_SIZE': 0
        }
    )

    while True:
        chunk = request.stream.read(8192)
        if not chunk:
            break
        parser.write(chunk)
    parser.finalize()
    parser.close()

    for file_name in file_names:
        params[file_name].close()

    return params


@ns_conf.route('/byom/<name>')
@ns_conf.param('name', "Name of the model")
class BYOMUpload(Resource):
    @ns_conf.doc('post_file')
    @api_endpoint_metrics('POST', '/handlers/byom/handler')
    def post(self, name):
        params = prepare_formdata()

        code_file_path = params['code'].name.decode()
        try:
            module_file_path = params['modules'].name.decode()
        except AttributeError:
            module_file_path = Path(code_file_path).parent / 'requirements.txt'
            module_file_path.touch()
            module_file_path = str(module_file_path)

        connection_args = {
            'code': code_file_path,
            'modules': module_file_path,
            'type': params.get('type')
        }

        session = SessionController()

        base_ml_handler = session.integration_controller.get_ml_handler(name)
        base_ml_handler.update_engine(connection_args)

        engine_storage = HandlerStorage(base_ml_handler.integration_id)

        engine_versions = [
            int(x) for x in engine_storage.get_connection_args()['versions'].keys()
        ]

        return {
            'last_engine_version': max(engine_versions),
            'engine_versions': engine_versions
        }

    @ns_conf.doc('put_file')
    @api_endpoint_metrics('PUT', '/handlers/byom/handler')
    def put(self, name):
        ''' upload new model
            params in FormData:
                - code
                - modules
        '''

        params = prepare_formdata()

        code_file_path = params['code'].name.decode()
        try:
            module_file_path = params['modules'].name.decode()
        except KeyError:
            module_file_path = Path(code_file_path).parent / 'requirements.txt'
            module_file_path.touch()
            module_file_path = str(module_file_path)

        connection_args = {
            'code': code_file_path,
            'modules': module_file_path,
            'mode': params.get('mode'),
            'type': params.get('type')
        }

        ast_query = CreateMLEngine(
            name=Identifier(name),
            handler='byom',
            params=connection_args
        )
        sql_session = SessionController()
        command_executor = ExecuteCommands(sql_session)
        command_executor.execute_command(ast_query)

        return '', 200



import io
import pandas as pd
import dropbox

from dropbox.exceptions import AuthError, ApiError, BadInputError
from typing import Dict, Optional, Text

from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb_sql_parser.ast import Select, Identifier, Insert

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)


from mindsdb.integrations.libs.api_handler import APIHandler, APIResource


class ListFilesTable(APIResource):

    def list(self, conditions=None, limit=None, sort=None, targets=None, **kwargs):
        files = self.handler._list_files()
        data = []
        for file in files:
            item = {
                "path": file["path"],
                "name": file["name"],
                "extension": file["extension"],
            }
            data.append(item)
        df = pd.DataFrame(data)
        return df

    def get_columns(self):
        return ["path", "name", "extension"]


class FileTable(APIResource):

    def _get_file_df(self):
        try:
            df = self.handler._read_file(self.table_name)
            if df is None:
                raise Exception(f"No such file found for the path: {self.dropbox_path}")

            return df
        except Exception as e:
            self.handler.logger.error(e)

    def list(self, conditions=None, limit=None, sort=None, targets=None, **kwargs):
        return self._get_file_df()

    def get_columns(self):
        df = self.handler._read_file(self.table_name)
        return df.columns.tolist()

    def insert(self, query: Insert) -> None:
        columns = [col.name for col in query.columns]
        data = [dict(zip(columns, row)) for row in query.values]
        df_new = pd.DataFrame(data)
        df_existing = self._get_file_df()
        df_combined = pd.concat([df_existing, df_new], ignore_index=True)
        self.handler._write_file(self.table_name, df_combined)


class DropboxHandler(APIHandler):

    name = "dropbox"
    supported_file_formats = ["csv", "tsv", "json", "parquet"]

    def __init__(self, name: Text, connection_data: Optional[Dict], **kwargs):
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.logger = log.getLogger(__name__)
        self.dbx = None
        self.is_connected = False
        self._files_table = ListFilesTable(self)
        self._register_table("files", self._files_table)

    def connect(self):
        try:
            if self.is_connected:
                return
            if "access_token" not in self.connection_data:
                raise ValueError("Access token must be provided.")
            self.dbx = dropbox.Dropbox(self.connection_data["access_token"])
            self.is_connected = True
            self.logger.info(
                f"Connected to Dropbox as {self.dbx.users_get_current_account().email}"
            )
        except ValueError as e:
            self.logger.error(f"Error connecting to Dropbox: {e}")
        except AuthError as e:
            self.logger.error(f"Authentication error with Dropbox: {e}")
        except BadInputError as e:
            self.logger.error(f"Bad input error with Dropbox: {e}")
        except Exception as e:
            self.logger.error(f"Error with Dropbox: {e}")

    def check_connection(self) -> StatusResponse:
        response = StatusResponse(False)
        try:
            self.connect()
            response.success = True
        except (ApiError, ValueError) as e:
            self.logger.error(f"Error connecting to Dropbox with Dropbox: {e}")
            response.error_message = str(e)
        except AuthError as e:
            self.logger.error(f"Authentication error with Dropbox: {e}")
            response.error_message = str(e)
        except Exception as e:
            self.logger.error(f"Error with Dropbox Handler: {e}")
            response.error_message = str(e)
        return response

    def disconnect(self):
        if not self.is_connected:
            return
        self.dbx = None
        self.is_connected = False
        self.logger.info("Disconnected from Dropbox")

    def _read_as_content(self, file_path) -> None:
        """
        Read files as content
        """
        try:
            _, res = self.dbx.files_download(file_path)
            content = res.content
            return content
        except ApiError as e:
            self.logger.error(f"Error when downloading a file from Dropbox: {e}")

    def query(self, query: ASTNode) -> Response:

        if isinstance(query, Select):
            table_name = query.from_table.parts[-1]
            if table_name == "files":
                table = self._files_table
                df = table.select(query)

                # add content
                has_content = False
                for target in query.targets:
                    if (
                        isinstance(target, Identifier)
                        and target.parts[-1].lower() == "content"
                    ):
                        has_content = True
                        break
                if has_content:
                    df["content"] = df["path"].apply(self._read_as_content)
            else:
                table = FileTable(self, table_name=table_name)
                df = table.select(query)

            return Response(RESPONSE_TYPE.TABLE, data_frame=df)
        elif isinstance(query, Insert):
            table_name = query.table.parts[-1]
            table = FileTable(self, table_name=table_name)
            table.insert(query)
            return Response(RESPONSE_TYPE.OK)
        else:
            raise NotImplementedError(
                "Only SELECT and INSERT operations are supported."
            )

    def get_tables(self) -> Response:
        table_names = list(self._tables.keys())
        df = pd.DataFrame(table_names, columns=["table_name"])
        return Response(RESPONSE_TYPE.TABLE, data_frame=df)

    def get_columns(self, table_name: str) -> Response:
        table = self._get_table(Identifier(table_name))
        columns = table.get_columns()
        df = pd.DataFrame(columns, columns=["column_name"])
        return Response(RESPONSE_TYPE.TABLE, data_frame=df)

    def _list_files(self, path=""):
        files = []
        result = self.dbx.files_list_folder(path, recursive=True)
        files.extend(self._process_entries(result.entries))
        while result.has_more:
            result = self.dbx.files_list_folder_continue(result.cursor)
            files.extend(self._process_entries(result.entries))
        return files

    def _process_entries(self, entries):
        files = []
        for entry in entries:
            if isinstance(entry, dropbox.files.FileMetadata):
                extension = entry.name.split(".")[-1].lower()
                if extension in self.supported_file_formats:
                    files.append(
                        {
                            "path": entry.path_lower,
                            "name": entry.name,
                            "extension": extension,
                        }
                    )
        return files

    def _read_file(self, path) -> pd.DataFrame:
        try:
            _, res = self.dbx.files_download(path)
            content = res.content
            extension = path.split(".")[-1].lower()
            if extension == "csv":
                df = pd.read_csv(io.BytesIO(content))
            elif extension == "tsv":
                df = pd.read_csv(io.BytesIO(content), sep="\t")
            elif extension == "json":
                df = pd.read_json(io.BytesIO(content))
            elif extension == "parquet":
                df = pd.read_parquet(io.BytesIO(content))
            else:
                raise ValueError(f"Unsupported file format: {extension}")
            return df
        except ValueError as e:
            self.logger.error(f"Error with file extension: {e}")
        except ApiError as e:
            self.logger.error(f"Error when downloading a file from Dropbox: {e}")
        except Exception as e:
            self.logger.error(f"Error with Dropbox Handler: {e}")

    def _write_file(self, path, df: pd.DataFrame):
        try:
            extension = path.split(".")[-1].lower()
            buffer = io.BytesIO()
            if extension == "csv":
                df.to_csv(buffer, index=False)
            elif extension == "tsv":
                df.to_csv(buffer, index=False, sep="\t")
            elif extension == "json":
                df.to_json(buffer, orient="records")
            elif extension == "parquet":
                df.to_parquet(buffer, index=False)
            else:
                raise ValueError(f"Unsupported file format: {extension}")
            buffer.seek(0)
            self.dbx.files_upload(
                buffer.read(), path, mode=dropbox.files.WriteMode.overwrite
            )
        except ValueError as e:
            self.logger.error(f"Error with file extension: {e}")
        except ApiError as e:
            self.logger.error(f"Error when writing a file to Dropbox: {e}")
        except Exception as e:
            self.logger.error(f"Error with Dropbox Handler: {e}")



from mindsdb.integrations.handlers.scylla_handler import Handler as ScyllaHandler
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response)


class CassandraHandler(ScyllaHandler):
    """
    This handler handles connection and execution of the Cassandra statements.
    """

    name = 'cassandra'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

    def get_tables(self) -> Response:
        """
        Get the list of tables in the connected Cassandra database.

        :return: List of table names.
        """
        sql = "DESCRIBE TABLES"
        result = self.native_query(sql)
        df = result.data_frame
        df = df.rename(columns={'name': 'table_name'})
        result.data_frame = df
        return result



import os
import copy
from typing import Dict, List, Optional

import pandas as pd
import hashlib
import numpy as np

from mindsdb_sql_parser.ast import (
    BinaryOperation,
    Constant,
    Identifier,
    Select,
    Update,
    Delete,
    Star
)
from mindsdb_sql_parser.ast.mindsdb import CreatePredictor

from mindsdb.integrations.utilities.query_traversal import query_traversal

import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.libs.vectordatabase_handler import (
    DistanceFunction,
    TableField,
    VectorStoreHandler,
)
from mindsdb.integrations.utilities.rag.rag_pipeline_builder import RAG
from mindsdb.integrations.utilities.rag.config_loader import load_rag_config
from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.integrations.handlers.langchain_embedding_handler.langchain_embedding_handler import construct_model_from_args, row_to_document

from mindsdb.interfaces.agents.constants import DEFAULT_EMBEDDINGS_MODEL_CLASS
from mindsdb.interfaces.agents.langchain_agent import create_chat_model, get_llm_provider
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.knowledge_base.preprocessing.models import PreprocessingConfig, Document
from mindsdb.interfaces.knowledge_base.preprocessing.document_preprocessor import PreprocessorFactory
from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator
from mindsdb.utilities.context import context as ctx

from mindsdb.api.executor.command_executor import ExecuteCommands
from mindsdb.utilities import log
from mindsdb.integrations.utilities.rag.rerankers.reranker_compressor import LLMReranker

logger = log.getLogger(__name__)

KB_TO_VECTORDB_COLUMNS = {
    'id': 'original_row_id',
    'chunk_id': 'id',
    'chunk_content': 'content'
}


def get_embedding_model_from_params(embedding_model_params: dict):
    """
    Create embedding model from parameters.
    """
    params_copy = copy.deepcopy(embedding_model_params)
    provider = params_copy.pop('provider', None).lower()
    api_key = get_api_key(provider, params_copy, strict=False) or params_copy.get('api_key')
    # Underscores are replaced because the provider name ultimately gets mapped to a class name.
    # This is mostly to support Azure OpenAI (azure_openai); the mapped class name is 'AzureOpenAIEmbeddings'.
    params_copy['class'] = provider.replace('_', '')
    if provider == 'azure_openai':
        # Azure OpenAI expects the api_key to be passed as 'openai_api_key'.
        params_copy['openai_api_key'] = api_key
    else:
        params_copy[f"{provider}_api_key"] = api_key
    params_copy.pop('api_key', None)
    params_copy['model'] = params_copy.pop('model_name', None)

    return construct_model_from_args(params_copy)


def get_reranking_model_from_params(reranking_model_params: dict):
    """
    Create reranking model from parameters.
    """
    params_copy = copy.deepcopy(reranking_model_params)
    provider = params_copy.pop('provider', "openai").lower()
    if provider != 'openai':
        raise ValueError("Only OpenAI provider is supported for the reranking model.")
    params_copy[f"{provider}_api_key"] = get_api_key(provider, params_copy, strict=False) or params_copy.get('api_key')
    params_copy.pop('api_key', None)
    params_copy['model'] = params_copy.pop('model_name', None)

    return LLMReranker(**params_copy)


class KnowledgeBaseTable:
    """
    Knowledge base table interface
    Handlers requests to KB table and modifies data in linked vector db table
    """

    def __init__(self, kb: db.KnowledgeBase, session):
        self._kb = kb
        self._vector_db = None
        self.session = session
        self.document_preprocessor = None
        self.document_loader = None
        self.model_params = None

    def configure_preprocessing(self, config: Optional[dict] = None):
        """Configure preprocessing for the knowledge base table"""
        logger.debug(f"Configuring preprocessing with config: {config}")
        self.document_preprocessor = None  # Reset existing preprocessor
        if config is not None:
            preprocessing_config = PreprocessingConfig(**config)
            self.document_preprocessor = PreprocessorFactory.create_preprocessor(preprocessing_config)
            logger.debug(f"Created preprocessor of type: {type(self.document_preprocessor)}")
        else:
            # Always create a default preprocessor if none specified
            self.document_preprocessor = PreprocessorFactory.create_preprocessor()
            logger.debug("Created default preprocessor")

    def select_query(self, query: Select) -> pd.DataFrame:
        """
        Handles select from KB table.
        Replaces content values with embeddings in where clause. Sends query to vector db
        :param query: query to KB table
        :return: dataframe with the result table
        """
        logger.debug(f"Processing select query: {query}")

        # Extract the content query text for potential reranking

        db_handler = self.get_vector_db()

        logger.debug("Replaced content with embeddings in where clause")
        # set table name
        query.from_table = Identifier(parts=[self._kb.vector_database_table])
        logger.debug(f"Set table name to: {self._kb.vector_database_table}")

        requested_kb_columns = []
        for target in query.targets:
            if isinstance(target, Star):
                requested_kb_columns = None
                break
            else:
                requested_kb_columns.append(target.parts[-1].lower())

        query.targets = [
            Identifier(TableField.ID.value),
            Identifier(TableField.CONTENT.value),
            Identifier(TableField.METADATA.value),
            Identifier(TableField.DISTANCE.value),
        ]

        # Get response from vector db
        logger.debug(f"Using vector db handler: {type(db_handler)}")

        # extract values from conditions and prepare for vectordb
        conditions = []
        query_text = None
        reranking_threshold = None
        query_conditions = db_handler.extract_conditions(query.where)
        if query_conditions is not None:
            for item in query_conditions:
                if item.column == "reranking_threshold" and item.op.value == "=":
                    try:
                        reranking_threshold = float(item.value)
                        # Validate range: must be between 0 and 1
                        if not (0 <= reranking_threshold <= 1):
                            raise ValueError(f"reranking_threshold must be between 0 and 1, got: {reranking_threshold}")
                        logger.debug(f"Found reranking_threshold in query: {reranking_threshold}")
                    except (ValueError, TypeError) as e:
                        error_msg = f"Invalid reranking_threshold value: {item.value}. {str(e)}"
                        logger.error(error_msg)
                        raise ValueError(error_msg)
                elif item.column == TableField.CONTENT.value:
                    query_text = item.value

                    # replace content with embeddings
                    conditions.append(FilterCondition(
                        column=TableField.EMBEDDINGS.value,
                        value=self._content_to_embeddings(item.value),
                        op=FilterOperator.EQUAL,
                    ))
                else:
                    conditions.append(item)

        logger.debug(f"Extracted query text: {query_text}")

        self.addapt_conditions_columns(conditions)
        df = db_handler.dispatch_select(query, conditions)
        df = self.addapt_result_columns(df)

        logger.debug(f"Query returned {len(df)} rows")
        logger.debug(f"Columns in response: {df.columns.tolist()}")
        # Check if we have a rerank_model configured in KB params

        df = self.add_relevance(df, query_text, reranking_threshold)

        # filter by targets
        if requested_kb_columns is not None:
            df = df[requested_kb_columns]
        return df

    def add_relevance(self, df, query_text, reranking_threshold=None):
        relevance_column = TableField.RELEVANCE.value

        reranking_model_params = self._kb.params.get("reranking_model")
        if reranking_model_params and query_text and len(df) > 0:
            # Use reranker for relevance score
            try:
                logger.info(f"Using knowledge reranking model from params: {reranking_model_params}")
                # Apply custom filtering threshold if provided
                if reranking_threshold is not None:
                    reranking_model_params["filtering_threshold"] = reranking_threshold
                    logger.info(f"Using custom filtering threshold: {reranking_threshold}")

                reranker = get_reranking_model_from_params(reranking_model_params)
                # Get documents to rerank
                documents = df['chunk_content'].tolist()
                # Use the get_scores method with disable_events=True
                scores = reranker.get_scores(query_text, documents)
                # Add scores as the relevance column
                df[relevance_column] = scores

                # Filter by threshold
                scores_array = np.array(scores)
                df = df[scores_array > reranker.filtering_threshold]
                logger.debug(f"Applied reranking with params: {reranking_model_params}")
            except Exception as e:
                logger.error(f"Error during reranking: {str(e)}")
                # Fallback to distance-based relevance
                if 'distance' in df.columns:
                    df[relevance_column] = 1 / (1 + df['distance'])
                else:
                    logger.info("No distance or reranker available")

        elif 'distance' in df.columns:
            # Calculate relevance from distance
            logger.info("Calculating relevance from vector distance")
            df[relevance_column] = 1 / (1 + df['distance'])
            if reranking_threshold is not None:
                df = df[df[relevance_column] > reranking_threshold]

        else:
            df[relevance_column] = None
            df['distance'] = None
        # Sort by relevance
        df = df.sort_values(by=relevance_column, ascending=False)
        return df

    def addapt_conditions_columns(self, conditions):
        if conditions is None:
            return
        for condition in conditions:
            if condition.column in KB_TO_VECTORDB_COLUMNS:
                condition.column = KB_TO_VECTORDB_COLUMNS[condition.column]

    def addapt_result_columns(self, df):
        col_update = {}
        for kb_col, vec_col in KB_TO_VECTORDB_COLUMNS.items():
            if vec_col in df.columns:
                col_update[vec_col] = kb_col

        df = df.rename(columns=col_update)

        columns = list(df.columns)
        # update id, get from metadata
        df[TableField.ID.value] = df[TableField.METADATA.value].apply(
            lambda m: None if m is None else m.get('original_row_id')
        )

        # id on first place
        return df[[TableField.ID.value] + columns]

    def insert_files(self, file_names: List[str]):
        """Process and insert files"""
        if not self.document_loader:
            raise ValueError("Document loader not configured")

        documents = list(self.document_loader.load_files(file_names))
        if documents:
            self.insert_documents(documents)

    def insert_web_pages(
            self,
            urls: List[str],
            crawl_depth: int,
            limit: int,
            filters: List[str] = None
    ):
        """Process and insert web pages"""
        if not self.document_loader:
            raise ValueError("Document loader not configured")

        documents = list(self.document_loader.load_web_pages(
            urls,
            limit=limit,
            crawl_depth=crawl_depth,
            filters=filters
        ))
        if documents:
            self.insert_documents(documents)

    def insert_query_result(self, query: str, project_name: str):
        """Process and insert SQL query results"""
        if not self.document_loader:
            raise ValueError("Document loader not configured")

        documents = list(self.document_loader.load_query_result(query, project_name))
        if documents:
            self.insert_documents(documents)

    def insert_rows(self, rows: List[Dict]):
        """Process and insert raw data rows"""
        if not rows:
            return

        documents = [Document(
            content=row.get('content', ''),
            id=row.get('id'),
            metadata=row.get('metadata', {})
        ) for row in rows]

        self.insert_documents(documents)

    def insert_documents(self, documents: List[Document]):
        """Process and insert documents with preprocessing if configured"""
        df = pd.DataFrame([doc.model_dump() for doc in documents])

        self.insert(df)

    def update_query(self, query: Update):
        # add embeddings to content in updated collumns
        query = copy.deepcopy(query)

        emb_col = TableField.EMBEDDINGS.value
        cont_col = TableField.CONTENT.value
        if cont_col in query.update_columns:
            content = query.update_columns[cont_col]

            # Apply preprocessing to content if configured
            if self.document_preprocessor:
                doc = Document(
                    content=content.value,
                    metadata={}  # Empty metadata for content-only updates
                )
                processed_chunks = self.document_preprocessor.process_documents([doc])
                if processed_chunks:
                    content.value = processed_chunks[0].content

            query.update_columns[emb_col] = Constant(self._content_to_embeddings(content))

        # TODO search content in where clause?

        # set table name
        query.table = Identifier(parts=[self._kb.vector_database_table])

        # send to vectordb
        db_handler = self.get_vector_db()
        conditions = db_handler.extract_conditions(query.where)
        self.addapt_conditions_columns(conditions)
        db_handler.dispatch_update(query, conditions)

    def delete_query(self, query: Delete):
        """
        Handles delete query to KB table.
        Replaces content values with embeddings in WHERE clause. Sends query to vector db
        :param query: query to KB table
        """
        query_traversal(query.where, self._replace_query_content)

        # set table name
        query.table = Identifier(parts=[self._kb.vector_database_table])

        # send to vectordb
        db_handler = self.get_vector_db()
        conditions = db_handler.extract_conditions(query.where)
        self.addapt_conditions_columns(conditions)
        db_handler.dispatch_delete(query, conditions)

    def hybrid_search(
        self,
        query: str,
        keywords: List[str] = None,
        metadata: Dict[str, str] = None,
        distance_function=DistanceFunction.COSINE_DISTANCE
    ) -> pd.DataFrame:
        query_df = pd.DataFrame.from_records([{TableField.CONTENT.value: query}])
        embeddings_df = self._df_to_embeddings(query_df)
        if embeddings_df.empty:
            return pd.DataFrame([])
        embeddings = embeddings_df.iloc[0][TableField.EMBEDDINGS.value]
        keywords_query = None
        if keywords is not None:
            keywords_query = ' '.join(keywords)
        db_handler = self.get_vector_db()
        return db_handler.hybrid_search(
            self._kb.vector_database_table,
            embeddings,
            query=keywords_query,
            metadata=metadata,
            distance_function=distance_function
        )

    def clear(self):
        """
        Clear data in KB table
        Sends delete to vector db table
        """
        db_handler = self.get_vector_db()
        db_handler.delete(self._kb.vector_database_table)

    def insert(self, df: pd.DataFrame):
        """Insert dataframe to KB table."""
        if df.empty:
            return

        try:
            run_query_id = ctx.run_query_id
            # Link current KB to running query (where KB is used to insert data)
            if run_query_id is not None:
                self._kb.query_id = run_query_id
                db.session.commit()

        except AttributeError:
            ...

        # First adapt column names to identify content and metadata columns
        adapted_df = self._adapt_column_names(df)
        content_columns = self._kb.params.get('content_columns', [TableField.CONTENT.value])

        # Convert DataFrame rows to documents, creating separate documents for each content column
        raw_documents = []
        for idx, row in adapted_df.iterrows():
            base_metadata = self._parse_metadata(row.get(TableField.METADATA.value, {}))
            provided_id = row.get(TableField.ID.value)

            for col in content_columns:
                content = row.get(col)
                if content and str(content).strip():
                    content_str = str(content)

                    # Use provided_id directly if it exists, otherwise generate one
                    doc_id = self._generate_document_id(content_str, col, provided_id)

                    # Need provided ID to link chunks back to original source (e.g. database row).
                    row_id = provided_id if provided_id else idx

                    metadata = {
                        **base_metadata,
                        'original_row_id': str(row_id),
                        'content_column': col,
                    }

                    raw_documents.append(Document(
                        content=content_str,
                        id=doc_id,
                        metadata=metadata
                    ))

        # Apply preprocessing to all documents if preprocessor exists
        if self.document_preprocessor:
            processed_chunks = self.document_preprocessor.process_documents(raw_documents)
        else:
            processed_chunks = raw_documents  # Use raw documents if no preprocessing

        # Convert processed chunks back to DataFrame with standard structure
        df = pd.DataFrame([{
            TableField.CONTENT.value: chunk.content,
            TableField.ID.value: chunk.id,
            TableField.METADATA.value: chunk.metadata
        } for chunk in processed_chunks])

        if df.empty:
            logger.warning("No valid content found in any content columns")
            return

        # add embeddings and send to vector db
        df_emb = self._df_to_embeddings(df)
        df = pd.concat([df, df_emb], axis=1)
        db_handler = self.get_vector_db()
        db_handler.do_upsert(self._kb.vector_database_table, df)

    def _adapt_column_names(self, df: pd.DataFrame) -> pd.DataFrame:
        '''
        Convert input columns for vector db input
        - id, content and metadata
        '''
        # Debug incoming data
        logger.debug(f"Input DataFrame columns: {df.columns}")
        logger.debug(f"Input DataFrame first row: {df.iloc[0].to_dict()}")

        params = self._kb.params
        columns = list(df.columns)

        # -- prepare id --
        id_column = params.get('id_column')
        if id_column is not None and id_column not in columns:
            id_column = None

        if id_column is None and TableField.ID.value in columns:
            id_column = TableField.ID.value

        # Also check for case-insensitive 'id' column
        if id_column is None:
            column_map = {col.lower(): col for col in columns}
            if 'id' in column_map:
                id_column = column_map['id']

        if id_column is not None:
            columns.remove(id_column)
            logger.debug(f"Using ID column: {id_column}")

        # Create output dataframe
        df_out = pd.DataFrame()

        # Add ID if present
        if id_column is not None:
            df_out[TableField.ID.value] = df[id_column]
            logger.debug(f"Added IDs: {df_out[TableField.ID.value].tolist()}")

        # -- prepare content and metadata --
        content_columns = params.get('content_columns', [TableField.CONTENT.value])
        metadata_columns = params.get('metadata_columns')

        logger.debug(f"Processing with: content_columns={content_columns}, metadata_columns={metadata_columns}")

        # Handle SQL query result columns
        if content_columns:
            # Ensure content columns are case-insensitive
            column_map = {col.lower(): col for col in columns}
            content_columns = [
                column_map.get(col.lower(), col)
                for col in content_columns
            ]
            logger.debug(f"Mapped content columns: {content_columns}")

        if metadata_columns:
            # Ensure metadata columns are case-insensitive
            column_map = {col.lower(): col for col in columns}
            metadata_columns = [
                column_map.get(col.lower(), col)
                for col in metadata_columns
            ]
            logger.debug(f"Mapped metadata columns: {metadata_columns}")

        if content_columns is not None:
            content_columns = list(set(content_columns).intersection(columns))
            if len(content_columns) == 0:
                raise ValueError(f'Content columns {params.get("content_columns")} not found in dataset: {columns}')

            if metadata_columns is not None:
                metadata_columns = list(set(metadata_columns).intersection(columns))
            else:
                # all the rest columns
                metadata_columns = list(set(columns).difference(content_columns))

        # Add content columns directly (don't combine them)
        for col in content_columns:
            df_out[col] = df[col]

        # Add metadata
        if metadata_columns and len(metadata_columns) > 0:
            def convert_row_to_metadata(row):
                metadata = {}
                for col in metadata_columns:
                    value = row[col]
                    # Convert numpy/pandas types to Python native types
                    if pd.api.types.is_datetime64_any_dtype(value) or isinstance(value, pd.Timestamp):
                        value = str(value)
                    elif pd.api.types.is_integer_dtype(value):
                        value = int(value)
                    elif pd.api.types.is_float_dtype(value):
                        value = float(value)
                    elif pd.api.types.is_bool_dtype(value):
                        value = bool(value)
                    elif isinstance(value, dict):
                        metadata.update(value)
                        continue
                    else:
                        value = str(value)
                    metadata[col] = value
                return metadata

            metadata_dict = df[metadata_columns].apply(convert_row_to_metadata, axis=1)
            df_out[TableField.METADATA.value] = metadata_dict

        logger.debug(f"Output DataFrame columns: {df_out.columns}")
        logger.debug(f"Output DataFrame first row: {df_out.iloc[0].to_dict() if not df_out.empty else 'Empty'}")

        return df_out

    def _replace_query_content(self, node, **kwargs):
        if isinstance(node, BinaryOperation):
            if isinstance(node.args[0], Identifier) and isinstance(node.args[1], Constant):
                col_name = node.args[0].parts[-1]
                if col_name.lower() == TableField.CONTENT.value:
                    # replace
                    node.args[0].parts = [TableField.EMBEDDINGS.value]
                    node.args[1].value = [self._content_to_embeddings(node.args[1].value)]

    def get_vector_db(self) -> VectorStoreHandler:
        """
        helper to get vector db handler
        """
        if self._vector_db is None:
            database = db.Integration.query.get(self._kb.vector_database_id)
            if database is None:
                raise ValueError('Vector database not found. Is it deleted?')
            database_name = database.name
            self._vector_db = self.session.integration_controller.get_data_handler(database_name)
        return self._vector_db

    def get_vector_db_table_name(self) -> str:
        """
        helper to get underlying table name used for embeddings
        """
        return self._kb.vector_database_table

    def _df_to_embeddings(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Returns embeddings for input dataframe.
        Uses model embedding model to convert content to embeddings.
        Automatically detects input and output of model using model description
        :param df:
        :return: dataframe with embeddings
        """

        if df.empty:
            return pd.DataFrame([], columns=[TableField.EMBEDDINGS.value])

        # keep only content
        df = df[[TableField.CONTENT.value]]

        model_id = self._kb.embedding_model_id
        if model_id:
            # get the input columns
            model_rec = db.session.query(db.Predictor).filter_by(id=model_id).first()

            assert model_rec is not None, f"Model not found: {model_id}"
            model_project = db.session.query(db.Project).filter_by(id=model_rec.project_id).first()

            project_datanode = self.session.datahub.get(model_project.name)

            model_using = model_rec.learn_args.get('using', {})
            input_col = model_using.get('question_column')
            if input_col is None:
                input_col = model_using.get('input_column')

            if input_col is not None and input_col != TableField.CONTENT.value:
                df = df.rename(columns={TableField.CONTENT.value: input_col})

            df_out = project_datanode.predict(
                model_name=model_rec.name,
                df=df,
                params=self.model_params
            )

            target = model_rec.to_predict[0]
            if target != TableField.EMBEDDINGS.value:
                # adapt output for vectordb
                df_out = df_out.rename(columns={target: TableField.EMBEDDINGS.value})

        elif self._kb.params.get('embedding_model'):
            embedding_model = get_embedding_model_from_params(self._kb.params.get('embedding_model'))

            df_texts = df.apply(row_to_document, axis=1)
            embeddings = embedding_model.embed_documents(df_texts.tolist())
            df_out = df.copy().assign(**{TableField.EMBEDDINGS.value: embeddings})

        else:
            raise ValueError("No embedding model found for the knowledge base.")

        df_out = df_out[[TableField.EMBEDDINGS.value]]

        return df_out

    def _content_to_embeddings(self, content: str) -> List[float]:
        """
        Converts string to embeddings
        :param content: input string
        :return: embeddings
        """
        df = pd.DataFrame([[content]], columns=[TableField.CONTENT.value])
        res = self._df_to_embeddings(df)
        return res[TableField.EMBEDDINGS.value][0]

    def build_rag_pipeline(self, retrieval_config: dict):
        """
        Builds a RAG pipeline with returned sources

        Args:
            retrieval_config: dict with retrieval config

        Returns:
            RAG: Configured RAG pipeline instance

        Raises:
            ValueError: If the configuration is invalid or required components are missing
        """
        # Get embedding model from knowledge base
        embeddings_model = None
        if self._kb.embedding_model:
            # Extract embedding model args from knowledge base table
            embedding_args = self._kb.embedding_model.learn_args.get('using', {})
            # Construct the embedding model directly
            embeddings_model = construct_model_from_args(embedding_args)
            logger.debug(f"Using knowledge base embedding model with args: {embedding_args}")
        elif self._kb.params.get('embedding_model'):
            embeddings_model = get_embedding_model_from_params(self._kb.params['embedding_model'])
            logger.debug(f"Using knowledge base embedding model from params: {self._kb.params['embedding_model']}")
        else:
            embeddings_model = DEFAULT_EMBEDDINGS_MODEL_CLASS()
            logger.debug("Using default embedding model as knowledge base has no embedding model")

        # Update retrieval config with knowledge base parameters
        kb_params = {
            'vector_store_config': {
                'kb_table': self
            }
        }

        # Load and validate config
        try:
            rag_config = load_rag_config(retrieval_config, kb_params, embeddings_model)

            # Build LLM if specified
            if 'llm_model_name' in rag_config:
                llm_args = {"model_name": rag_config.llm_model_name}
                if not rag_config.llm_provider:
                    llm_args['provider'] = get_llm_provider(llm_args)
                else:
                    llm_args["provider"] = rag_config.llm_provider
                rag_config.llm = create_chat_model(llm_args)

            # Create RAG pipeline
            rag = RAG(rag_config)
            logger.debug(f"RAG pipeline created with config: {rag_config}")
            return rag

        except Exception as e:
            logger.error(f"Error building RAG pipeline: {str(e)}")
            raise ValueError(f"Failed to build RAG pipeline: {str(e)}")

    def _parse_metadata(self, base_metadata):
        """Helper function to robustly parse metadata string to dict"""
        if isinstance(base_metadata, dict):
            return base_metadata
        if isinstance(base_metadata, str):
            try:
                import ast
                return ast.literal_eval(base_metadata)
            except (SyntaxError, ValueError):
                logger.warning(f"Could not parse metadata: {base_metadata}. Using empty dict.")
                return {}
        return {}

    def _generate_document_id(self, content: str, content_column: str, provided_id: str = None) -> str:
        """
        Generate a deterministic document ID from content and column name.
        If provided_id exists, combines it with content_column.

        Args:
            content: The content string
            content_column: Name of the content column
            provided_id: Optional user-provided ID
        Returns:
            Deterministic document ID
        """
        if provided_id is not None:
            return f"{provided_id}_{content_column}"

        id_string = f"content={content}_column={content_column}"
        return hashlib.sha256(id_string.encode()).hexdigest()

    def _convert_metadata_value(self, value):
        """
        Convert metadata value to appropriate Python type.

        Args:
            value: The value to convert

        Returns:
            Converted value in appropriate Python type
        """
        if pd.isna(value):
            return None

        # Handle pandas/numpy types
        if pd.api.types.is_datetime64_any_dtype(value) or isinstance(value, pd.Timestamp):
            return str(value)
        elif pd.api.types.is_integer_dtype(type(value)):
            return int(value)
        elif pd.api.types.is_float_dtype(type(value)):
            return float(value)
        elif pd.api.types.is_bool_dtype(type(value)):
            return bool(value)

        # Handle basic Python types
        if isinstance(value, (int, float, bool)):
            return value

        # Convert everything else to string
        return str(value)


class KnowledgeBaseController:
    """
    Knowledge base controller handles all
    manages knowledge bases
    """

    def __init__(self, session) -> None:
        self.session = session

    def add(
            self,
            name: str,
            project_name: str,
            embedding_model: Identifier,
            storage: Identifier,
            params: dict,
            preprocessing_config: Optional[dict] = None,
            if_not_exists: bool = False
    ) -> db.KnowledgeBase:
        """
        Add a new knowledge base to the database
        :param preprocessing_config: Optional preprocessing configuration to validate and store
        :param is_sparse: Whether to use sparse vectors for embeddings
        :param vector_size: Optional size specification for vectors, required when is_sparse=True
        """
        # Validate preprocessing config first if provided
        if preprocessing_config is not None:
            PreprocessingConfig(**preprocessing_config)  # Validate before storing
            params = params or {}
            params['preprocessing'] = preprocessing_config

        # Check if vector_size is provided when using sparse vectors
        is_sparse = params.get('is_sparse')
        vector_size = params.get('vector_size')
        if is_sparse and vector_size is None:
            raise ValueError("vector_size is required when is_sparse=True")

        # get project id
        project = self.session.database_controller.get_project(project_name)
        project_id = project.id

        # not difference between cases in sql
        name = name.lower()
        # check if knowledge base already exists
        kb = self.get(name, project_id)
        if kb is not None:
            if if_not_exists:
                return kb
            raise EntityExistsError("Knowledge base already exists", name)

        embedding_model_params = params.get('embedding_model', None)
        reranking_model_params = params.get('reranking_model', None)

        if embedding_model:
            model_name = embedding_model.parts[-1]

        elif embedding_model_params:
            # Get embedding model from params.
            # This is called here to check validaity of the parameters.
            get_embedding_model_from_params(
                embedding_model_params
            )

        else:
            model_name = self._get_default_embedding_model(
                project.name,
                params=params
            )
            params['default_embedding_model'] = model_name

        model_project = None
        if embedding_model is not None and len(embedding_model.parts) > 1:
            # model project is set
            model_project = self.session.database_controller.get_project(embedding_model.parts[-2])
        elif not embedding_model_params:
            model_project = project

        embedding_model_id = None
        if model_project:
            model = self.session.model_controller.get_model(
                name=model_name,
                project_name=model_project.name
            )
            model_record = db.Predictor.query.get(model['id'])
            embedding_model_id = model_record.id

        if reranking_model_params:
            # Get reranking model from params.
            # This is called here to check validaity of the parameters.
            get_reranking_model_from_params(reranking_model_params)

        # search for the vector database table
        if storage is None:
            cloud_pg_vector = os.environ.get('KB_PGVECTOR_URL')
            if cloud_pg_vector:
                vector_table_name = name
                # Add sparse vector support for pgvector
                vector_db_params = {}
                # Check both explicit parameter and model configuration
                is_sparse = is_sparse or model_record.learn_args.get('using', {}).get('sparse')
                if is_sparse:
                    vector_db_params['is_sparse'] = True
                    if vector_size is not None:
                        vector_db_params['vector_size'] = vector_size
                vector_db_name = self._create_persistent_pgvector(vector_db_params)

            else:
                # create chroma db with same name
                vector_table_name = "default_collection"
                vector_db_name = self._create_persistent_chroma(name)
                # memorize to remove it later
                params['default_vector_storage'] = vector_db_name
        elif len(storage.parts) != 2:
            raise ValueError('Storage param has to be vector db with table')
        else:
            vector_db_name, vector_table_name = storage.parts

        # create table in vectordb before creating KB
        self.session.datahub.get(vector_db_name).integration_handler.create_table(
            vector_table_name
        )
        vector_database_id = self.session.integration_controller.get(vector_db_name)['id']

        # Store sparse vector settings in params if specified
        if is_sparse:
            params = params or {}
            params['vector_config'] = {
                'is_sparse': is_sparse
            }
            if vector_size is not None:
                params['vector_config']['vector_size'] = vector_size

        kb = db.KnowledgeBase(
            name=name,
            project_id=project_id,
            vector_database_id=vector_database_id,
            vector_database_table=vector_table_name,
            embedding_model_id=embedding_model_id,
            params=params,
        )
        db.session.add(kb)
        db.session.commit()
        return kb

    def _create_persistent_pgvector(self, params=None):
        """Create default vector database for knowledge base, if not specified"""
        vector_store_name = "kb_pgvector_store"

        # check if exists
        if self.session.integration_controller.get(vector_store_name):
            return vector_store_name

        self.session.integration_controller.add(vector_store_name, 'pgvector', params or {})
        return vector_store_name

    def _create_persistent_chroma(self, kb_name, engine="chromadb"):
        """Create default vector database for knowledge base, if not specified"""

        vector_store_name = f"{kb_name}_{engine}"

        vector_store_folder_name = f"{vector_store_name}"
        connection_args = {"persist_directory": vector_store_folder_name}

        # check if exists
        if self.session.integration_controller.get(vector_store_name):
            return vector_store_name

        self.session.integration_controller.add(vector_store_name, engine, connection_args)
        return vector_store_name

    def _get_default_embedding_model(self, project_name, engine="langchain_embedding", params: dict = None):
        """create a default embedding model for knowledge base, if not specified"""
        model_name = "kb_default_embedding_model"

        # check exists
        try:
            model = self.session.model_controller.get_model(model_name, project_name=project_name)
            if model is not None:
                return model_name
        except PredictorRecordNotFound:
            pass

        using_args = {
            'engine': engine
        }
        if engine == 'langchain_embedding':
            # Use default embeddings.
            using_args['class'] = 'openai'

        # Include API key if provided.
        using_args.update({k: v for k, v in params.items() if 'api_key' in k})
        statement = CreatePredictor(
            name=Identifier(parts=[project_name, model_name]),
            using=using_args,
            targets=[
                Identifier(parts=[TableField.EMBEDDINGS.value])
            ]
        )

        command_executor = ExecuteCommands(self.session)
        command_executor.answer_create_predictor(statement, project_name)

        return model_name

    def delete(self, name: str, project_name: int, if_exists: bool = False) -> None:
        """
        Delete a knowledge base from the database
        """
        try:
            project = self.session.database_controller.get_project(project_name)
        except ValueError:
            raise ValueError(f"Project not found: {project_name}")
        project_id = project.id

        # check if knowledge base exists
        kb = self.get(name, project_id)
        if kb is None:
            # knowledge base does not exist
            if if_exists:
                return
            else:
                raise EntityNotExistsError("Knowledge base does not exist", name)

        # kb exists
        db.session.delete(kb)
        db.session.commit()

        # drop objects if they were created automatically
        if 'default_vector_storage' in kb.params:
            try:
                handler = self.session.datahub.get(kb.params['default_vector_storage']).integration_handler
                handler.drop_table(kb.vector_database_table)
                self.session.integration_controller.delete(kb.params['default_vector_storage'])
            except EntityNotExistsError:
                pass
        if 'default_embedding_model' in kb.params:
            try:
                self.session.model_controller.delete_model(kb.params['default_embedding_model'], project_name)
            except EntityNotExistsError:
                pass

    def get(self, name: str, project_id: int) -> db.KnowledgeBase:
        """
        Get a knowledge base from the database
        by name + project_id
        """
        kb = (
            db.session.query(db.KnowledgeBase)
            .filter_by(
                name=name,
                project_id=project_id,
            )
            .first()
        )
        return kb

    def get_table(self, name: str, project_id: int, params: dict = None) -> KnowledgeBaseTable:
        """
        Returns kb table object with properly configured preprocessing
        :param name: table name
        :param project_id: project id
        :param params: runtime parameters for KB. Keys: 'model' - parameters for embedding model
        :return: kb table object
        """
        kb = self.get(name, project_id)
        if kb is not None:
            table = KnowledgeBaseTable(kb, self.session)
            if params:
                table.model_params = params.get('model')

            # Always configure preprocessing - either from params or default
            if kb.params and 'preprocessing' in kb.params:
                table.configure_preprocessing(kb.params['preprocessing'])
            else:
                table.configure_preprocessing(None)  # This ensures default preprocessor is created

            return table

    def list(self, project_name: str = None) -> List[dict]:
        """
        List all knowledge bases from the database
        belonging to a project
        """
        project_controller = ProjectController()
        projects = project_controller.get_list()
        if project_name is not None:
            projects = [p for p in projects if p.name == project_name]

        query = (
            db.session.query(db.KnowledgeBase)
            .filter(db.KnowledgeBase.project_id.in_(list([p.id for p in projects])))
        )

        data = []
        project_names = {
            i.id: i.name
            for i in project_controller.get_list()
        }

        for record in query:
            vector_database = record.vector_database
            embedding_model = record.embedding_model

            data.append({
                'id': record.id,
                'name': record.name,
                'project_id': record.project_id,
                'project_name': project_names[record.project_id],
                'embedding_model': embedding_model.name if embedding_model is not None else None,
                'vector_database': None if vector_database is None else vector_database.name,
                'vector_database_table': record.vector_database_table,
                'query_id': record.query_id,
                'params': record.params
            })

        return data

    def update(self, name: str, project_id: int, **kwargs) -> db.KnowledgeBase:
        """
        Update a knowledge base record
        """
        raise NotImplementedError()



import queue
from typing import Any, Dict, List, Optional, Sequence
from uuid import UUID

from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.documents import Document


class EventDispatchCallbackHandler(BaseCallbackHandler):
    '''Puts dispatched events onto an event queue to be processed as a streaming chunk'''
    def __init__(self, queue: queue.Queue):
        self.queue = queue

    def on_custom_event(
        self,
        name: str,
        data: Any,
        *,
        run_id: UUID,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        self.queue.put({
            'type': 'event',
            'name': name,
            'data': data
        })

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        document_objects = []
        for d in documents:
            document_objects.append({
                'content': d.page_content,
                'metadata': d.metadata
            })
        self.queue.put({
            'type': 'event',
            'name': 'retriever_end',
            'data': {
                'documents': document_objects
            }
        })
